#
# License-Identifier: GPL
#
# Copyright (C) 2024 The Yambo Team
#
# Authors: HPC, FP
#
# This file is part of the yambopy project
#
import os
import numpy as np
from netCDF4 import Dataset
from yambopy.tools.jsonencoder import JsonDumper, JsonLoader
from yambopy.lattice import vol_lat, rec_lat, car_red
from yambopy.kpoints import expand_kpoints,make_kpositive
from yambopy.tools.string import marquee

class YamboLatticeDB(object):
    """
    Class to read the lattice information from the netcdf file
    """
    def __init__(self,lat=None,alat=None,sym_car=None,iku_kpoints=None,
                      car_atomic_positions=None,atomic_numbers=None,
                      time_rev=None,nelectrons=None,spinor_components=None,
                      mag_syms=None):
        self.lat                  = np.array(lat)
        self.alat                 = np.array(alat)
        self.sym_car              = np.array(sym_car)
        self.iku_kpoints          = np.array(iku_kpoints)
        self.car_atomic_positions = np.array(car_atomic_positions)
        self.atomic_numbers       = np.array(atomic_numbers)
        self.time_rev             = time_rev
        self.spinor_components    = spinor_components
        self.nelectrons           = nelectrons
        self.mag_syms             = mag_syms
        self.ibz_nkpoints         = len(iku_kpoints)

    @classmethod
    def from_db(cls,filename='ns.db1',Expand=True,atol=1e-6):
        return cls.from_db_file(filename,Expand,atol)
    
    @classmethod
    def from_db_file(cls,filename='ns.db1',Expand=True,atol=1e-6):
        """ Initialize YamboLattice from a local dbfile """

        if not os.path.isfile(filename):
            raise FileNotFoundError("error opening %s in YamboLatticeDB"%filename)

        with Dataset(filename) as database:

            dimensions = database.variables['DIMENSIONS'][:]
            time_rev   = dimensions[9].astype(int)
            spinor_components = dimensions[11].astype(int)
            nelectrons = dimensions[14].astype(int)
            mag_syms   = database.variables['mag_syms'][:].astype(int)[0]

            natoms_a = database.variables['N_ATOMS'][:].astype(int).T
            tmp_an = database.variables['atomic_numbers'][:].astype(int)
            tmp_apos = database.variables['ATOM_POS'][:,:]
            # Prevent case where n. of atomic species > n. of species in ATOM_POS
            i_at_unused_types = [i_at for i_at, at_type_n in enumerate(natoms_a) if at_type_n == 0]
            if len(i_at_unused_types)>0:
                print('[WARNING] Found unused atomic type(s): ',i_at_unused_types)
                natoms_a = [at_type_n for at_type_n in natoms_a if at_type_n!=0]
                tmp_an = [tmp_an[i] for i in range(len(tmp_an)) if tmp_an[i] not in i_at_unused_types ]

            flatten = lambda l: [item for sublist in l for item in sublist]
            atomic_numbers = flatten([[tmp_an[n]]*na for n,na in enumerate(natoms_a)])
            atomic_positions = np.vstack([[tmp_apos[n,ia] for ia in range(na)] for n,na in enumerate(natoms_a) ])

            args = dict( atomic_numbers       = atomic_numbers,
                         car_atomic_positions = atomic_positions,
                         sym_car              = np.transpose( database.variables['SYMMETRY'][:], (0,2,1) ), # transpose leaving first axis as symm index
                         iku_kpoints          = database.variables['K-POINTS'][:].T,
                         lat                  = database.variables['LATTICE_VECTORS'][:].T,
                         alat                 = database.variables['LATTICE_PARAMETER'][:].T,
                         time_rev             = time_rev,
                         spinor_components    = spinor_components,
                         nelectrons           = nelectrons,
                         mag_syms             = mag_syms)

        y = cls(**args)
        if Expand: y.expand_kpoints(atol=atol,verbose=0)
        return y
 
    @classmethod    
    def from_dict(cls,data):
        """ Initialize instance of the class from a dictionary
        """
        lat = data["lat"]
        alat = data["alat"]
        sym_car = data["sym_car"]
        iku_kpoints = data["iku_kpoints"]
        atomic_positions = data["car_atomic_positions"]
        atomic_numbers = data["atomic_numbers"]
        time_rev = data["time_rev"]

        y = cls(lat,alat,sym_car,iku_kpoints,atomic_positions,atomic_numbers,time_rev)
        return y

    @classmethod
    def from_json_file(cls,filename):
        data = JsonLoader(filename)
        return cls.from_dict(data)
 
    @property
    def nkpoints(self):
        return len(self.car_kpoints)

    @property
    def red_atomic_positions(self):
        return car_red(self.car_atomic_positions,self.lat)

    def as_dict(self):
        """ get the information of this class as a dictionary
        """
        data = {"lat" : self.lat,
                "alat" : self.alat,
                "sym_car" : self.sym_car,
                "iku_kpoints" : self.iku_kpoints,
                "car_atomic_positions" : self.car_atomic_positions,
                "atomic_numbers" : self.atomic_numbers,
                "time_rev": self.time_rev }
        return data

    def write_json(self,filename):
        """ write a json file with the lattice information """
        JsonDumper(self.as_dict(),filename)

    @property
    def iku_kpoints(self):
        return self._iku_kpoints

    @iku_kpoints.setter
    def iku_kpoints(self,value):
        if hasattr(self,"_red_kpoints"): delattr(self,"_red_kpoints")
        if hasattr(self,"_car_kpoints"): delattr(self,"_car_kpoints")
        self._iku_kpoints = value

    @property
    def nkpoints(self):
        return len(self.iku_kpoints)

    @property
    def nsym(self):
        return len(self.sym_car)

    @property
    def rlat(self):
        """calculate the reciprocal lattice"""
        if not hasattr(self,'_rlat'):
            self._rlat = rec_lat(self.lat)
        return self._rlat 

    @property
    def rlat_vol(self):
        return (2*np.pi)**3 * vol_lat(self.rlat)

    @property
    def lat_vol(self):
        return vol_lat(self.lat)

    @property
    def car_kpoints(self):
        """convert form internal yambo units to cartesian lattice units"""
        if not hasattr(self,"_car_kpoints"):
            self._car_kpoints = np.array([ k/self.alat for k in self.iku_kpoints ])
        return self._car_kpoints

    @property
    def red_kpoints(self):
        """convert from cartesian coordinates to reduced coordinates"""
        if not hasattr(self,"_red_kpoints"):
            self._red_kpoints = car_red(self.car_kpoints,self.rlat)
        return self._red_kpoints

    @property
    def k_grid(self,atol=1.e-6):
        """Return the k-points grid dimensions """
        if hasattr(self,"_kgrid"):
            return self._kgrid
        if not hasattr(self,"_red_kpoints"):
            self._red_kpoints = car_red(self.car_kpoints,self.rlat)
        self._kgrid =np.zeros(3,dtype=int)
        self._small_q=np.full(3,1e4,dtype=float)
        for idx in range(3): 
            for kpt in self._red_kpoints:
                val = abs(kpt[idx])
                if atol < val < self._small_q[idx]:
                    self._small_q[idx]=val
        self._red_kpoints=make_kpositive(self._red_kpoints)
        for idx in range(3): 
            for kpt in self._red_kpoints:
                n_grid = np.rint(kpt[idx]/self._small_q[idx])+1
                if self._kgrid[idx]< n_grid:
                    self._kgrid[idx]=n_grid
        return self._kgrid

    def get_ibz_kpoints(self,units='iku'):
        if hasattr(self,"ibz_kpoints"):
            kpts=self.ibz_kpoints
        else:
            kpts=self.iku_kpoints

        if units.lower()=='iku':
            return kpts
        elif units.lower()=='red':
            red_kpts=np.array([ k/self.alat for k in kpts ])
            return car_red(red_kpts,self.rlat)
        elif units.lower()=='car':
            return np.array([ k/self.alat for k in kpts ])
        else:
            raise Exception("Sorry, wrong k-points units, use: iku, red or car") 


  
    @property
    def sym_red(self):
        """Convert cartesian transformations to reduced transformations"""
        if not hasattr(self,"_sym_red"):
            sym_red = np.zeros([self.nsym,3,3],dtype=int)
            for n,s in enumerate(self.sym_car):
                sym_red[n] = np.round(np.dot(np.dot(self.lat,s.T),np.linalg.inv(self.lat)))
            self._sym_red = sym_red
        return self._sym_red

    @property
    def sym_rec_red(self):
        """Convert reduced transformations to reduced reciprocal transformations"""
        if not hasattr(self,"_sym_rec_red"):
            sym_rec_red = np.zeros([self.nsym,3,3],dtype=int)
            for n,s in enumerate(self.sym_red):
                sym_rec_red[n] = np.linalg.inv(s).T
            self._sym_rec_red = sym_rec_red
        return self._sym_rec_red
         
    @property
    def sym_rec(self):
        """Convert cartesian transformations to reciprocal transformations"""
        if not hasattr(self,"_sym_rec"):
            sym_rec = np.zeros([self.nsym,3,3])
            for n,s in enumerate(self.sym_car):
                sym_rec[n] = np.linalg.inv(s).T
            self._sym_rec = sym_rec
        return self._sym_rec

    @property
    def time_rev_list(self):
        """get a list of symmetries with time reversal"""
        time_rev_list = [False]*self.nsym
        for i in range(self.nsym):
            time_rev_list[i] = ( i >= self.nsym/(self.time_rev+1) )
        return time_rev_list

    @property
    def nbandsv(self):
        """
        number of occupied bands
        """
        if (self.spinor_components==2): return int(self.nelectrons)
        else:                           return int(self.nelectrons/2)

    def expand_kpoints(self,verbose=0,atol=1.e-6):
        """
        Wrapper for expand_kpoints in kpoints module

        Take a list of qpoints and symmetry operations and return the full brillouin zone
        with the corresponding index in the irreducible brillouin zone
        """

        # Store original kpoints in iku coordinates
        self.ibz_kpoints = self.iku_kpoints

        weights, BZ_to_IBZ_indexes, symmetry_indexes, kpoints_full = expand_kpoints(self.car_kpoints,self.sym_car,self.rlat,atol=atol)

        if verbose: print("%d kpoints expanded to %d"%(len(self.car_kpoints),len(kpoints_full)))

        #set the variables
        self.weights_ibz      = weights
        self.symmetry_indexes = symmetry_indexes
        self.iku_kpoints      = [k*self.alat for k in kpoints_full]
        # BZ_to_IBZ_indexes[ik_bz] = ik_ibz in the original unexpanded list
        self.BZ_to_IBZ_indexes = BZ_to_IBZ_indexes # for clarity for users
        self.kpoints_indexes   = BZ_to_IBZ_indexes # for compatibility
        # IBZ_to_BZ_indexes[ik_ibz] = {ik_bz} in the star (first element is identity)
        self.IBZ_to_BZ_indexes = {}
        for ibz_index in np.unique(BZ_to_IBZ_indexes):
            self.IBZ_to_BZ_indexes[ibz_index] = np.where(BZ_to_IBZ_indexes == ibz_index)[0]
        kmap = np.zeros((self.nkpoints, 2), dtype=int)
        kmap[:, 0] = self.kpoints_indexes
        kmap[:, 1] = self.symmetry_indexes
        self.kmap  = kmap

    def get_units_info(self):

        info_string = \
        "          bohr^-1 | Yambo cartesian units [cc] in yambo: \n\
                ::   self.car_kpoints*2.*pi\n\
         \n\
          QE cartesian units [cart. coord. in units 2pi/alat] in QE: \n\
                ::   self.car_kpoints*self.alat[0]\n\
         \n\
          Internal yambo units [iku]: \n\
                ::   self.iku_kpoints\n\
         \n\
          Reduced coordinates [rlu in yambo, cryst. coord. in QE]: \n\
                ::   self.red_kpoints\n"
        print(info_string)

    def __str__(self):
        lines = []; app = lines.append
        app(marquee(self.__class__.__name__))
        app("lattice:")
        lines += [("%12.8lf " * 3) % tuple(vec) for vec in self.lat]
        app("atom positions:")
        for an, pos in zip(self.atomic_numbers, self.red_atomic_positions):
            app( "%3d " % an + ("%12.8lf " * 3) % tuple(pos) )
        if self.ibz_nkpoints!=self.nkpoints: app(f"{self.ibz_nkpoints} kpoints expanded to {self.nkpoints}")
        else: app(f"{self.nkpoints} kpoints in the IBZ")
        app(f"Time-reversal symmetry: {bool(self.time_rev)}")
        return "\n".join(lines)



#
# License-Identifier: GPL
#
# Copyright (C) 2024 The Yambo Team
#
# Authors: HPC
#
# This file is part of the yambopy project
#
def marquee(string,width=60,mark='='):
    return string.center(width).replace(" ",mark)



"""
Module that manages the parsing of the ``ndb.RT_carriers`` database created by `yambo_rt`.
"""
from netCDF4 import Dataset
from yambopy.units import ha2ev
from yambopy.plot.spectra import get_spectra

import numpy as np

class YamboRT_Carriers_DB():
    """
    Class to manage information about the real time distrubtion of carriers from the
    ``ndb.RT_carriers`` database created by `yambo_rt`.

    Args:
        file (:py:class:`string`) : string with the name of the database to be parsed
        verbose (:py:class:`boolean`) : define the amount of information provided on terminal

    Attributes:
        E_bare (:py:class:`np.array`) : Array that contains the bare bands energies.
            The structure of the array is [numk*numbands]. Energies are expressed in eV
            The structure is k1*b1,k1*b2,..,k1*bn,k2*b1,...
        f_bare (:py:class:`np.array`) : Array that contains the bare bands occupations.
            The structure of the array is [numk*numbands]
        kpoints (:py:class:`np.array`) : Array that contains the kpoints in cartesian  coordinates
            in units of 2*np.pi/alat (with a vector alat). The structure of the array is [numk,3],
            the first index runs over the kpoints and the second one gives the component
        bands_kpts (:py:class:`np.array`) : Array with the [first_band,last_band,numk]
        k_weight (:py:class:`np.array`) : Array that contains the weights of the kpoints.
            The structure of the array is [numk]
        delta_E (:py:class:`np.array`) : Array that contains the time-dependent correction
            to the bands energies. The structure of the array is [time,numk*numbands].
            Energies are expressed in eV
        delta_f (:py:class:`np.array`) : Array that contains the time-dependent corrections to
            the bands occupations. The structure of the array is [time,numk*numbands]

    """

    def __init__(self,folder='.',calc='SAVE',carriers_db='ndb.RT_carriers'):
        # Find path with RT data
        self.carriers_path = '%s/%s/%s'%(folder,calc,carriers_db)
        self.calc=calc
        try:
            data_obs= Dataset(self.carriers_path)
        except:
            raise ValueError("Error reading CARRIERS database at %s"%self.carriers_path)

        self.readDB(data_obs)

        data_obs.close()


    def readDB(self,database):
        """
        Read the data from the ``ndb.RT_carriers`` database created by `yambo_rt`. The variables
        are extracted from the database and stored in the attributes of the object.

        Args:
            verbose (:py:class:`boolean`) : define the amount of information provided on terminal

        """

        self.E_bare = ha2ev*np.array(database.variables['RT_carriers_E_bare'])
        self.f_bare = np.array(database.variables['RT_carriers_f_bare'])
        self.kpoints = np.array(database.variables['RT_kpt'][:].T)
        self.bands_kpts = np.array(database.variables['RT_bands_kpts'])
        self.k_weight = np.array(database.variables['RT_k_weight'])
        self.delta_E = ha2ev*np.array(database.variables['RT_carriers_delta_E'])
        self.delta_f = np.hstack(np.array(database.variables['RT_carriers_delta_f']))

    def get_info(self):
        """
        Provide information on the attributes of the class
        """
        print('YamboRTCarriersParser variables structure')
        print('Bands used and number of k-points',self.bands_kpts)
        print('kpoints shape',self.kpoints.shape)
        print('E_bare shape',self.E_bare.shape)
        print('f_bare shape',self.f_bare.shape)
        print('delta_E shape',self.delta_E.shape)
        print('delta_f',self.delta_f.shape)


    def build_f_bare_dos(self, dE = 0.1, eta = 0.1, broad_kind = 'lorentzian'):
        """
        For each kpoint build a dos which expresses the bare occupation level in terms
        of the energy. The energy ranges from the minum to the maximum of the
        E_bare variable.

        Args:
            dE (:py:class:`float`) : energy step in eV
            eta (:py:class:`float`) : magnitude of the broading parameter (in the same units used for the values array)
            broad_kind (:py:class:`string`) : type of broading function used (lorentzian, gaussian)

        Returns:
            :py:class:`Dos` : Instance of the ``Dos`` class. The object is an array of dos, one for
                each kpoint

        """
        numbnds = self.bands_kpts[1]-self.bands_kpts[0]
        numkp = self.bands_kpts[2]
        Emin= min(self.E_bare)-10*eta
        Emax= max(max(self.E_bare*self.f_bare),max(self.E_bare*self.delta_f))+10*eta
        #
        # Here I want to calculate the weighted density of states (w-DOS) so I pass
        # to the get_spectra function the occupation as residual for the DOS
        #
        w, dos = get_spectra(energies=self.E_bare,residuals=self.f_bare,broadening=eta,emin=Emin,emax=Emax)
        return w,dos
    
    
    def build_delta_f_dos(self, dE = 0.1, eta = 0.1, broad_kind = 'lorentzian'):
        """
        For each kpoint build a dos which expresses the bare occupation level in terms
        of the energy. The energy ranges from the minum to the maximum of the
        E_bare variable.

        Args:
            dE (:py:class:`float`) : energy step in eV
            eta (:py:class:`float`) : magnitude of the broading parameter (in the same units used for the values array)
            broad_kind (:py:class:`string`) : type of broading function used (lorentzian, gaussian)

        Returns:
            :py:class:`Dos` : Instance of the ``Dos`` class. The object is an array of dos, one for
                each kpoint

        """
        numbnds = self.bands_kpts[1]-self.bands_kpts[0]
        numkp = self.bands_kpts[2]
        Emin= min(self.E_bare)-10*eta
        Emax= max(max(self.E_bare*self.f_bare),max(self.E_bare*self.delta_f))+10*eta
        #
        # Here I want to calculate the weighted density of states (w-DOS) so I pass
        # to the get_spectra function the occupation as residual for the DOS
        #
        w, dos = get_spectra(energies=self.E_bare,residuals=self.delta_f,broadening=eta,emin=Emin,emax=Emax)
        return w,dos





# License-Identifier: GPL
#
# Copyright (C) 2024 The Yambo Team
#
# Authors: MRF
#
# This file is part of the yambopy project
#
import numpy as np
from yambopy import YamboLatticeDB,YamboExcitonDB,YamboDipolesDB
from yambopy.units import kb,autime2s,ha2ev,m_e,speed_of_light
import os
from yambopy.tools.citations import citation

def get_exciton_dipole(state, blongdir, ylat, ydip, yexc, _cache={}):
    """
    This function computes the dipole D of exciton state `state`
    at Q=0 starting from the transition-space expression:

    D_{state} = \sum_{cvk} A^{state}_{cvk} Efield \cdot r_cvk 

    :: A      --> BSE eigenvectors
    :: Efield --> electric field
    :: r      --> dipole matrix elements

    This is useful to change Efield direction without rerunning Yambo

    NB: in Yambo, q=0 is actually q=q0_def_norm=1e-5, so if you are checking for 
        consistency with the residuals in BS_diago db you have to multiply 
        by q0_def_norm

    Input:
    * blongdir:   electric field polarization direction
    * lattice_path, dipoles_path, bse_path : paths to required databases

    Output:
    * Values of |D|^2 in bohr^2
    """

    blongdir = np.asarray(blongdir, dtype=float)
    key = (id(ydip.dipoles), id(yexc.table), tuple(blongdir))

    # cache dipoles_proj and kcv_idx for same inputs
    if key not in _cache:
        field_dir = blongdir / np.linalg.norm(blongdir)
        dipoles_proj = np.einsum('x,kxcv->kcv', field_dir, ydip.dipoles)

        table_kcv = yexc.table[:, [0, 2, 1]]
        kcv_idx = tuple((table_kcv[:, :3] - 1).T)

        trans_dipoles = dipoles_proj[kcv_idx]
        _cache[key] = trans_dipoles
    else:
        trans_dipoles = _cache[key]

    dip_exc = np.dot(yexc.eigenvectors[state], trans_dipoles)
    dip_exc_squared = np.abs(dip_exc)**2

    return dip_exc_squared / ylat.nkpoints


@citation("H.-Y. Chen et al. Phys. Rev. B 100, 075135 (2019)")
def get_radiative_lifetime_3D_iso(T,state,ylat,ydip,yexc,Meff,eps,shiftE=0):

    """
        Function to compute the radiative lifetime tau_S(T) of a single exciton state for *isotropic* 3D materials.

        Input
            * T:        temperature in K
            * state:    index of the exciton state. Python convention (indices starting from 0)
            * ylat:     lattice database, YamboLatticeDB
            * ydip:     dipoles database, YamboDipolesDB
            * yexc:     BSE database, YamboExcitonDB
            * Meff:     exciton effective mass in electron rest mass (m_e) units. 
            * eps:      material's relative optical dielectric constant
            * shiftE:   rigid shift of exciton energies. Default: shift=0 eV

        Output
            * Radiative lifetime of exciton state in seconds
    """

    if not (isinstance(Meff,float) and isinstance(eps,float)):
        raise ValueError("Meff and eps must be floats")

    Omega = ylat.lat_vol  #Bohr**3

    muS2 = get_exciton_dipole(state,[1,1,1],ylat,ydip,yexc) # Bohr**2
    ES = np.real(yexc.eigenvalues[state])+shiftE  # eV

    Meff_eV = Meff*m_e   # eV
    
    T_factor = ( ES**2/(2*Meff_eV*kb*T) )**(3/2)
    dipole_factor = 8/3*np.sqrt(np.pi*eps)*muS2/Omega
    prefactor = 4*np.pi/autime2s
    
    return 1/(prefactor*dipole_factor*T_factor)

@citation("H.-Y. Chen et al. Phys. Rev. B 100, 075135 (2019)")
def get_radiative_lifetime_3D_aniso(T,state,ylat,ydip,yexc,Meff,eps,shiftE=0):
    """
        Function to compute the radiative lifetime tau_S(T) of a single exciton state for *anisotropic* (uniaxial) 3D materials.

        Input
            * T:        temperature in K
            * state:    index of the exciton state. Python convention (indices starting from 0)
            * ylat:     lattice database, YamboLatticeDB
            * ydip:     dipoles database, YamboDipolesDB
            * yexc:     BSE database, YamboExcitonDB
            * Meff:     2 element list: exciton effective masses in electron rest mass (m_e) units in-plane, out-of-plane: [Meff_xy, Meff_z]
            * eps:      2 element list: material's relative optical dielectric constants in-plane, out-of-plane: [eps_xy, eps_z]
            * shiftE:   rigid shift of exciton energies. Default: shift=0 eV

        Output
            * Radiative lifetime of exciton state in seconds
    """

    if not (isinstance(Meff,list) and len(Meff) == 2):
        raise ValueError("Meff must be lists of 2 floats: [Meff_xy, Meff_z]")
    if not (isinstance(eps,list) and len(eps) == 2):
        raise ValueError("eps must be lists of 2 floats: [eps_xy, eps_z]")

    Omega = ylat.lat_vol  #Bohr**3

    muS2_xy = get_exciton_dipole(state,[1,1,0],ylat,ydip,yexc) # Bohr**2
    muS2_z = get_exciton_dipole(state,[0,0,1],ylat,ydip,yexc)  # Bohr**2
    ES = np.real(yexc.eigenvalues[state])+shiftE  # eV

    Meff_xy_eV = Meff[0]*m_e  # eV
    Meff_z_eV = Meff[1]*m_e   # eV

    eps_xy = eps[0]
    eps_z = eps[1]

    T_factor = ( ES**2/(2*Meff_xy_eV**(2/3)*Meff_z_eV**(1/3)*kb*T) )**(3/2)
    dipole_factor = np.sqrt(np.pi*eps_xy)/Omega * ((2*eps_z/(3*eps_xy)+2)*muS2_xy+8/3*muS2_z)
    prefactor = 4*np.pi/autime2s
    
    return 1/(prefactor*dipole_factor*T_factor)  # seconds

@citation("H.-Y. Chen et al. Phys. Rev. B 100, 075135 (2019)")
def get_radiative_lifetime_2D(T,state,ylat,ydip,yexc,Meff,eps=1,dip_dir=[1,1,0],shiftE=0):
    """
        Function to compute the radiative lifetime tau_S(T) of a single exciton state for 2D materials.

        Input
            * T:        temperature in K
            * state:    index of the exciton state. Python convention (indices starting from 0)
            * ylat:     lattice database, YamboLatticeDB
            * ydip:     dipoles database, YamboDipolesDB
            * yexc:     BSE database, YamboExcitonDB
            * Meff:     exciton effective mass in electron rest mass (m_e) units. 
            * eps:      environment dielectric constant. Default: eps=1, vacuum.
            * dip_dir:  orientation of the dipole moments. Default: [1,1,0] in the xy plane.
            * shiftE:   rigid shift of exciton energies. Default: shift=0 eV

        Output
            * Radiative lifetime of exciton state in seconds
    """

    if not (isinstance(Meff,float) and isinstance(eps,float)):
        raise ValueError("Meff and eps must be floats")

    lat = ylat.lat
    A = np.cross(lat[0],lat[1])[2] # Bohr**2

    muS2 = get_exciton_dipole(state,dip_dir,ylat,ydip,yexc) # Bohr**2
    ES = np.real(yexc.eigenvalues[state])+shiftE # eV

    Meff_eV = Meff*m_e   # eV
    
    T_factor = 4/3*(ES**2/(2*Meff_eV*kb*T))
    gamma0_factor = (ES/ha2ev)*muS2/(eps*A)
    prefactor = 4*np.pi/speed_of_light/autime2s

    return 1/(prefactor*gamma0_factor*T_factor)   # seconds

@citation("H.-Y. Chen et al. Phys. Rev. B 100, 075135 (2019)")
def get_radiative_lifetime_1D(T,state,ylat,ydip,yexc,Meff,eps=1,dip_dir=[0,0,1],shiftE=0):
    """
        Function to compute the radiative lifetime tau_S(T) of a single exciton state for 1D materials.

        Input
            * T:        temperature in K
            * state:    index of the exciton state. Python convention (indices starting from 0)
            * ylat:     lattice database, YamboLatticeDB
            * ydip:     dipoles database, YamboDipolesDB
            * yexc:     BSE database, YamboExcitonDB
            * Meff:     exciton effective mass in electron rest mass (m_e) units. 
            * eps:      environment dielectric constant. Default: eps=1, vacuum.
            * dip_dir:  orientation of the dipole moments. Default: [0,0,1] along z.
            * shiftE:   rigid shift of exciton energies. Default: shift=0 eV

        Output
            * Radiative lifetime of exciton state in seconds
    """

    if not (isinstance(Meff,float) and isinstance(eps,float)):
        raise ValueError("Meff and eps must be floats")

    lat = ylat.lat
    Lz = lat[2,2]

    muS2 = get_exciton_dipole(state,dip_dir,ylat,ydip,yexc) # Bohr**2
    ES = np.real(yexc.eigenvalues[state])+shiftE  # eV

    Meff_eV = Meff*m_e   # eV
    
    T_factor = 4/3*(ES**2/(2*Meff_eV*kb*T))
    gamma0_factor = (ES/ha2ev)**2*muS2/(eps*Lz)
    prefactor = 4*np.pi/speed_of_light**2/autime2s
    
    return 1/(prefactor*gamma0_factor*T_factor)   # seconds

@citation("H.-Y. Chen et al. Phys. Rev. B 100, 075135 (2019)")
def get_radiative_lifetime_0D(state,ylat,ydip,yexc,eps=1,shiftE=0):
    """
        Function to compute the radiative lifetime tau_S(T) of a single exciton state for 0D materials.

        Input
            * state:    index of the exciton state. Python convention (indices starting from 0)
            * ylat:     lattice database, YamboLatticeDB
            * ydip:     dipoles database, YamboDipolesDB
            * yexc:     BSE database, YamboExcitonDB
            * eps:      environment dielectric constant. Default: eps=1, vacuum
            * shiftE:   rigid shift of exciton energies. Default: shift=0 eV.

        Output
            * Radiative lifetime of exciton state in seconds
    """

    if not isinstance(eps,float):
        raise ValueError("eps must be float")

    muS2 = get_exciton_dipole(state,[1,1,1],ylat,ydip,yexc) # Bohr**2
    ES = np.real(yexc.eigenvalues[state])+shiftE  # eV
    
    gamma0_factor = np.sqrt(eps)*(ES/ha2ev)**3*muS2
    prefactor = 4*np.pi/(3*speed_of_light**3)/autime2s
    
    return 1/(prefactor*gamma0_factor)   # seconds



def average_lifetime(Trange,states,ylat,ydip,yexc,dimension,Meff=None,eps=None,dip_dir=None,shiftE=0):
    """
        Function to compute the material's exciton radiative lifetime thermally averaged on the exciton states `states`.

        Input
            * Trange:       temperature range, in K
            * states:       list of indices of the exciton states to consider. Python convention (indices starting from 0.
            * ylat:         lattice database, YamboLatticeDB
            * ydip:         dipoles database, YamboDipolesDB
            * yexc:         BSE database, YamboExcitonDB
            * dimension:    dimensionality of the system. It can be:
                            - '3D' or '3D_iso' for isotropic 3D bulk systems
                            - '3D_aniso' for anisotropic (uniaxial) 3D bulks
                            - '2D', '1D', '0D' for lower-dimensional systems
            * Meff:         exciton effective mass. Needed by 3D, 2D, 1D calculations.
            * eps:          environment dielectric constant.
            * dip_dir:      dipole orientation. Needed by 2D and 1D calculations.
            * shiftE:   rigid shift of exciton energies. Default: shift=0 eV.

        Output
            * Array of averaged radiative lifetime with length as Trange in input
    """
    
    ES = np.real(yexc.eigenvalues)+shiftE
    DeltaE = ES-ES[states[0]]

    if dimension in ['3D','3D_iso']:
        gammas = np.array([np.exp(-DeltaE[state]/(kb*Trange))/get_radiative_lifetime_3D_iso(Trange,state,ylat,ydip,yexc,Meff,eps,shiftE) for state in states])
    elif dimension=='3D_aniso':
        gammas = np.array([np.exp(-DeltaE[state]/(kb*Trange))/get_radiative_lifetime_3D_aniso(Trange,state,ylat,ydip,yexc,Meff,eps,shiftE) for state in states])
    elif dimension=='2D':
        gammas = np.array([np.exp(-DeltaE[state]/(kb*Trange))/get_radiative_lifetime_2D(Trange,state,ylat,ydip,yexc,Meff,eps,dip_dir,shiftE) for state in states])
    elif dimension=='1D':
        gammas = np.array([np.exp(-DeltaE[state]/(kb*Trange))/get_radiative_lifetime_1D(Trange,state,ylat,ydip,yexc,Meff,eps,dip_dir,shiftE) for state in states])
    elif dimension=='0D':
        gammas = np.array([np.exp(-DeltaE[state]/(kb*Trange))/get_radiative_lifetime_0D(state,ylat,ydip,yexc,eps,shiftE) for state in states])
    else:
        raise ValueError("dimension must be one of the following: '3D','3D_iso','3D_aniso', '2D', '1D', '0D'")

    gamma_sum = np.sum(gammas,axis=0)

    normalize = np.array([np.exp(-DeltaE[state]/(kb*Trange)) for state in states])
    normalize_sum = np.sum(normalize,axis=0)

    return 1/(gamma_sum/normalize_sum)  # seconds




#
# Author: Henrique Pereira Coutada Miranda
# Run a Silicon groundstate calculation using Quantum Espresso
#
from __future__ import print_function, division
import sys
import argparse
from qepy import *
from schedulerpy import *
from math import sqrt

kpoints      = [6,6,1]
kpoints_nscf = [6,6,1]
kpoints_double = [24,24,1]
qpoints = [3,3,1]
layer_separation = 12
pw = 'pw.x'
ph = 'ph.x'
q2r = 'q2r.x'
matdyn = 'matdyn.x'
prefix = 'bn'

npoints = 10 
p = Path([ [[0.0, 0.0, 0.0],'$\Gamma$'],
           [[0.5, 0.0, 0.0],'M'],
           [[1./3,1./3,0.0],'K'],
           [[0.0, 0.0, 0.0],'$\Gamma$']], [int(npoints*2),int(npoints),int(sqrt(5)*npoints)])

# scheduler
scheduler = Scheduler.factory

# create the input files
def get_inputfile():
    """ Define a Quantum espresso input file for boron nitride
    """ 
    qe = PwIn()
    qe.set_atoms([['N',[0.0,0.0,0.5]],
                  ['B',[1/3,2/3,0.5]]])
    qe.atypes = {'B': [10.811, "B.pbe-mt_fhi.UPF"],
                 'N': [14.0067,"N.pbe-mt_fhi.UPF"]}

    qe.control['prefix'] = "'%s'"%prefix
    qe.control['verbosity'] = "'high'"
    qe.control['wf_collect'] = '.true.'
    qe.control['pseudo_dir'] = "'../pseudos/'"
    qe.system['celldm(1)'] = 4.7
    qe.system['celldm(3)'] = layer_separation/qe.system['celldm(1)']
    qe.system['ecutwfc'] = 60
    qe.system['occupations'] = "'fixed'"
    qe.system['nat'] = 2
    qe.system['ntyp'] = 2
    qe.system['ibrav'] = 4
    qe.kpoints = [9, 9, 1]
    qe.electrons['conv_thr'] = 1e-10
    return qe

#relax
def relax():
    if not os.path.isdir('relax'):
        os.mkdir('relax')
    qe = get_inputfile()
    qe.control['calculation'] = "'vc-relax'"
    qe.ions['ion_dynamics']  = "'bfgs'"
    qe.cell['cell_dynamics']  = "'bfgs'"
    qe.cell['cell_dofree']  = "'2Dxy'"
    qe.write('relax/%s.relax'%prefix)

#scf
def scf(kpoints,folder='scf'):
    if not os.path.isdir(folder):
        os.mkdir(folder)
    qe = get_inputfile()
    qe.control['calculation'] = "'scf'"
    qe.kpoints = kpoints
    qe.write('%s/%s.scf'%(folder,prefix))
 
#nscf
def nscf(kpoints,folder='nscf'):
    if not os.path.isdir(folder):
        os.mkdir(folder)
    qe = get_inputfile()
    qe.control['calculation'] = "'nscf'"
    qe.electrons['diago_full_acc'] = ".true."
    qe.electrons['conv_thr'] = 1e-8
    qe.system['nbnd'] = 70
    qe.system['force_symmorphic'] = ".true."
    qe.kpoints = kpoints
    qe.write('%s/%s.nscf'%(folder,prefix))

#bands
def bands():
    if not os.path.isdir('bands'):
        os.mkdir('bands')
    qe = get_inputfile()
    qe.control['calculation'] = "'bands'"
    qe.electrons['diago_full_acc'] = ".true."
    qe.electrons['conv_thr'] = 1e-6
    qe.system['nbnd'] = 6
    qe.system['force_symmorphic'] = ".true."
    qe.ktype = 'crystal'
    qe.set_path(p)
    qe.write('bands/%s.bands'%prefix)

def phonon(kpoints,qpoints,folder='phonon'):
    if not os.path.isdir(folder):
        os.mkdir(folder)
    ph = PhIn()
    ph['nq1'],ph['nq2'],ph['nq3'] = qpoints
    ph['tr2_ph'] = 1e-8
    ph['prefix'] = "'%s'"%prefix
    ph['epsil'] = ".false."
    ph['trans'] = ".true."
    ph['fildyn'] = "'%s.dyn'"%prefix
    ph['fildrho'] = "'%s.drho'"%prefix
    ph['ldisp'] = ".true."
    ph.write('%s/%s.ph'%(folder,prefix))

    md = DynmatIn()
    md['asr'] = "'simple'"
    md['fildyn'] = "'%s.dyn1'"%prefix
    md['filout'] = "'%s.modes'"%prefix
    md.write('%s/%s.dynmat'%(folder,prefix))

def update_positions(pathin,pathout):
    """ update the positions of the atoms in the scf file using the output of the relaxation loop
    """
    # Read scaled positions
    e = PwXML(prefix,path=pathin)
    pos = e.get_scaled_positions()
    
    #open relax input
    qin = PwIn.from_file('%s/%s.relax'%(pathin,prefix))
    print("old celldm(1)", qin.system['celldm(1)'])

    #open scf input
    qout = PwIn.from_file('%s/%s.scf'%(pathout,prefix))
    #replace lattice parameter
    qout.system['celldm(1)'] = e.cell[0][0]
    print("new celldm(1)", qout.system['celldm(1)'])
    #replace atomic positions
    new_atomic_pos = [[qout.atoms[i][0],list(pos[i])] for i in range(len(qout.atoms))]
    qout.set_atoms(new_atomic_pos)
    
    #re-write scf input
    qout.write('%s/%s.scf'%(pathout,prefix))

def run_plot():
    print("running plotting:")
    xml = PwXML(prefix=prefix,path='bands')
    xml.plot_eigen(p)

def run_projection(show=True):
    import matplotlib.pyplot as plt
    #write input file
    projwfc = ProjwfcIn('bn')
    projwfc.write(folder='bands')
    projwfc.run(folder='bands')
    #read xml file
    projection = ProjwfcXML(prefix='bn',path='bands')
    n_atom = range(16)
    b_atom = range(16,32)
    ax = plt.subplot(1,1,1)
    cax = projection.plot_eigen(ax,path=p,selected_orbitals=b_atom,selected_orbitals_2=n_atom,size=40,cmap='seismic')
    plt.colorbar(cax)
    if show: plt.show()

def run_bands(nthreads=1):
    print("running bands:")
    qe_run = scheduler() 
    qe_run.add_command("cp -r scf/%s.save bands/"%prefix)
    qe_run.add_command("cd bands; mpirun -np %d %s -inp %s.bands -nk %d > bands.log"%(nthreads,pw,prefix,nthreads))
    qe_run.run()
    qe_run.clean()
    print("done!")

if __name__ == "__main__":

    #parse options
    parser = argparse.ArgumentParser(description='Test the yambopy script.')
    parser.add_argument('-r' ,'--relax',       action="store_true", help='Structural relaxation')
    parser.add_argument('-s' ,'--scf',         action="store_true", help='Self-consistent calculation')
    parser.add_argument('-n' ,'--nscf',        action="store_true", help='Non-self consistent calculation')
    parser.add_argument('-n2','--nscf_double', action="store_true", help='Non-self consistent calculation for the double grid')
    parser.add_argument('-b' ,'--bands',       action="store_true", help='Calculate band-structure')
    parser.add_argument('-l' ,'--plot',        action="store_true", help='Plot band-structure')
    parser.add_argument('-o' ,'--orbitals',    action="store_true", help='Plot atomic orbital projected band-structure')
    parser.add_argument('-p' ,'--phonon',      action="store_true", help='Phonon calculation')
    parser.add_argument('-d' ,'--dispersion',  action="store_true", help='Phonon dispersion')
    parser.add_argument('-t' ,'--nthreads',                         help='Number of threads', default=2 )
    args = parser.parse_args()
    nthreads = int(args.nthreads)

    if len(sys.argv)==1:
        parser.print_help()
        sys.exit(1)

    # create input files and folders
    relax()
    scf(kpoints,folder='scf')
    nscf(kpoints_nscf)
    nscf(kpoints_double, folder='nscf_double')
    bands()
    phonon(kpoints,qpoints)

    if args.relax:
        print("running relax:")
        qe_run = scheduler() 
        qe_run.add_command("cd relax; mpirun -np %d %s -inp %s.relax > relax.log"%(nthreads,pw,prefix))  #relax
        qe_run.run()
        update_positions('relax','scf') 
        print("done!")

    if args.scf:
        print("running scf:")
        qe_run = scheduler() 
        qe_run.add_command("cd scf; mpirun -np %d %s -inp %s.scf > scf.log"%(nthreads,pw,prefix))  #scf
        qe_run.run()
        print("done!")
   
    if args.nscf: 
        print("running nscf:")
        qe_run = scheduler() 
        qe_run.add_command("cp -r scf/%s.save nscf/"%prefix) #nscf
        qe_run.add_command("cd nscf; mpirun -np %d %s -nk %d -inp %s.nscf > nscf.log"%(nthreads,pw,nthreads,prefix)) #nscf
        qe_run.run()
        print("done!")

    if args.nscf_double: 
        print("running nscf_double:")
        qe_run = scheduler() 
        qe_run.add_command("cp -r scf/%s.save nscf_double/"%prefix) #nscf
        qe_run.add_command("cd nscf_double; mpirun -np %d %s -inp %s.nscf > nscf_double.log"%(nthreads,pw,prefix)) #nscf
        qe_run.run()
        print("done!")
    
    if args.phonon:
        print("running phonon:")
        qe_run = scheduler() 
        qe_run.add_command("cp -r scf/%s.save phonon/"%prefix)
        qe_run.add_command("cd phonon; mpirun -np %d %s -inp %s.ph > phonon.log"%(nthreads,ph,prefix)) #phonon
        qe_run.add_command("dynmat.x < %s.dynmat > dynmat.log"%prefix) #matdyn
        qe_run.run()
        print("done!")

    if args.dispersion:
        qe_run = scheduler() 

        #q2r
        disp = DynmatIn()
        disp['fildyn']= "'%s.dyn'" % prefix
        disp['zasr']  = "'simple'"
        disp['flfrc'] = "'%s.fc'"  % prefix
        disp.write('phonon/q2r.in')
        qe_run.add_command('cd phonon; %s < q2r.in'%q2r)

        #dynmat
        dyn = DynmatIn()
        dyn['flfrc'] = "'%s.fc'" % prefix
        dyn['asr']   = "'simple'"
        dyn['flfrq'] = "'%s.freq'" % prefix
        dyn['q_in_cryst_coord'] = '.true.'
        dyn.qpoints = p.get_klist()
        dyn.write('phonon/matdyn.in')
        qe_run.add_command('%s < matdyn.in'%matdyn)
        qe_run.run()

        # matdyn class to read and plot the frequencies
        m = Matdyn.from_modes_file(folder='phonon')
        m.plot_eigen(path=p)
 
    if args.bands:
        run_bands(nthreads)
        run_plot()

    if args.orbitals:
        run_projection()

    if args.plot:
        run_plot()




# Copyright (C) 2018 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yamboparser
#
import os
import re
import numpy as np

#we try to use netcdf
try:
    from netCDF4 import Dataset
except ImportError:
    _has_netcdf = False
else:
    _has_netcdf = True

def if_has_netcdf(f):
    if _has_netcdf:
        return f
    
class YamboFile(object):
    """
    This is the Yambo file class.
    It takes as input a filename produced by yambo.
    Can be a netcdf or a text file

    List of supported NETCDF files:
        -> ndb.QP

    List of supported text files:
        -> r-*_em?1_*_gw0
        -> o-*.qp
    """
    _output_prefixes = ['o-']
    _report_prefixes = ['r-','r.']
    _log_prefixes    = ['l-','l.','l_']
    _netcdf_prefixes = ['ns','ndb']
    _netcdf_sufixes  = {'QP':'gw','HF_and_locXC':'hf'}
    _outputs_type = {'output_abs':'eps','output_loss':'eel','output_alpha':'alpha', 'output_jdos':'jdos'}

    def __init__(self,filename,folder='.',**parse_kwargs):
        self.filename = filename
        self.folder = folder
        self.errors = [] #list of errors
        self.warnings = [] #list of warnings
        self.memstats = [] #list of memory allocation statistics
        self.data = {} #dictionary containing all the important data from the file
        self.kpoints = {}
        self.timing = []

        #get the type of file
        self.type = YamboFile.get_filetype(filename,folder)

        #if needed read the lines
        if self.type in ['output_gw', 'log', 'report', 'output_abs', 'output_loss', 'output_alpha', 'output_jdos',]:
            #read lines from file
            with open(os.path.join(folder,filename),'r') as f:
                self.lines = f.readlines()
            
        #parse the file
        self.parse(**parse_kwargs)
    
    @staticmethod
    def get_filetype(filename,folder):
        """
        Get the type of file
        """
        type = 'unknown'
        basename = os.path.basename(filename)
        if any(basename.startswith(prefix) for prefix in YamboFile._output_prefixes):
            #read lines from file
            with open(os.path.join(folder,basename),'r') as f:
                lines = f.readlines()

            #get the line with the title
            title = lines[14]

            if 'GW' in title or '.qp' in filename:
                 type = 'output_gw'
            
            else:
                for key, val in list(zip(YamboFile._outputs_type.keys(),YamboFile._outputs_type.values())):
                    if val in filename: type = key

        elif any(basename.startswith(prefix) for prefix in YamboFile._report_prefixes):
            type = 'report'
        elif any(basename.startswith(prefix) for prefix in YamboFile._log_prefixes):
            type = 'log'
        elif any(basename.startswith(prefix) for prefix in YamboFile._netcdf_prefixes):
            for sufix in YamboFile._netcdf_sufixes:
                if basename.endswith(sufix):
                    type = 'netcdf_%s'%YamboFile._netcdf_sufixes[sufix]
                    break

        return type

    def parse(self,**parse_kwargs):
        """ Parse the file
            Add here things to read log and report files...
        """
        if   self.type == 'netcdf_gw': self.parse_netcdf_gw(**parse_kwargs)
        elif self.type == 'netcdf_hf': self.parse_netcdf_hf(**parse_kwargs)
        elif self.type in ['output_gw', 'output_abs', 'output_loss', 'output_alpha', 'output_jdos']: self.parse_output(**parse_kwargs)
        elif self.type == 'log': self.parse_log(**parse_kwargs)
        elif self.type == 'report'  : self.parse_report(**parse_kwargs)

    def parse_output(self,**parse_kwargs):
        """ Parse an output file from yambo,
        """
        zip_tags = parse_kwargs.get('zip_tags',False) #flag--default behavior is to do nothing
        #get the tags of the columns
        if self.type in YamboFile._outputs_type.keys():  #== "output_absorption":
            pattern = '([ `0-9a-zA-Z\-\/]+)\[[0-9]\]'
            tags = [tag.strip() for tag in re.findall(pattern,''.join(self.lines))]
            if len(tags) < 2 and self.type in ["output_loss", "output_abs", "output_alpha"]: # temporary fix for IP case: E[1] [eV]          Im(eps)            Re(eps)
                lines_with_matches = [line for line in self.lines if re.search(pattern, line)] 
                tags = [tag.strip() for tag in lines_with_matches[0].replace("#","").replace("\n","").replace("[eV]","").split()]
        if self.type == "output_gw":
            tags = [line.replace('(meV)','').replace('Sc(Eo)','Sc|Eo') for line in self.lines if all(tag in line for tag in ['K-point','Band','Eo'])][0]
            tags = tags[2:].strip().split()
        table = np.loadtxt(self.lines)
        _kdata ={}
        k_index =[ str(int(i)) for i in table[:,0]] # first column  has kpoints
        for ind in range(len(k_index)):
            for itag in range(len(tags)):
                 if k_index[ind] not in list(_kdata.keys()):
                     _kdata[k_index[ind]] = {}
                 try:
                     _kdata[k_index[ind]][tags[itag]].append(table[ind,itag]) #errors when you have multiple qp? IndexError: index 5 is out of bounds for axis 1 with size 5
                 except KeyError:
                     _kdata[k_index[ind]][tags[itag]]  = [ table[ind,itag] ]

        self.data = _kdata
        if (zip_tags): #combines tags such that keys refer to the columns in data file
            self.data = dict(zip(tags,table.T))

    @if_has_netcdf
    def parse_netcdf_gw(self,**parse_kwargs):
        """ Parse the netcdf gw file
        """
        data = {}

        filename = '%s/%s'%(self.folder,self.filename)
        with Dataset(filename) as f:

            #quasiparticles table
            qp_table  = f.variables['QP_table'][:]
            data['Kpoint_index'] = qp_table[2]
            data['Band'] = qp_table[0]
            #print(qp_table.shape)
            if qp_table.shape[0] == 4: # spin polarized
                data['Spin_pol'] = qp_table[3]
            data['qp_table'] = qp_table[:]  # ib, ik, ,(isp if spin polarized)
            #qpoints
            data['Kpoint']   = f.variables['QP_kpts'][:].T

            #quasi-particles
            if 'QP_E_Eo_Z' in f.variables:
                #old format
                qp = f.variables['QP_E_Eo_Z'][:]
                qp = qp[0]+qp[1]*1j
                data['E'],  data['Eo'], data['Z'] = qp.T
                data['E-Eo'] = data['E']  -  data['Eo']
                self.data=data
            else:
                #new format
                E  = f.variables['QP_E'][:]
                data['E'] = E[:,0] + E[:,1]*1j
                Eo = f.variables['QP_Eo'][:]
                data['Eo']= np.array(Eo,dtype=data['E'].dtype)
                Z  = f.variables['QP_Z'][:]
                data['Z'] = Z[:,0] + Z[:,1]*1j
                data['E-Eo'] = data['E']  -  data['Eo']
                self.data=data

    @if_has_netcdf
    def parse_netcdf_hf(self,**parse_kwargs):
        """ Parse the netcdf hf file (ndb.HF_and_locXC)
        """
        data = {}

        f = Dataset(os.path.join(self.folder,self.filename))

        pars = f.variables['PARS'][:]

        data['nkpoints'] = int(pars[0])
        data['nbands'] = int(pars[1])
        #data['QP_table'] = f.variables['QP_table'][:]

        #old format
        if 'Sx_Vxc' in f.variables:
            hf = f.variables['Sx_Vxc'][:]
            if hf.shape[0]%8 ==0 :
                qp =  hf.reshape(-1,8)
                ib, ibp, ik, isp, rsx, isx, revx, imvx = qp.T
            else:
                qp =  hf.reshape(-1,7)
                ib, ibp, ik, rsx, isx, revx, imvx = qp.T
            data['Sx'] = rsx + isx*1j
            data['Vxc'] = revx + imvx*1j
        #new format
        else:
            Sx  = f.variables['Sx'][:]
            data['Sx'] = Sx[:,0] + Sx[:,1]*1j
            Vxc = f.variables['Vxc'][:]
            data['Vxc'] = Vxc[:,0] + Vxc[:,1]*1j
            
        self.data=data
        f.close()

    def parse_report(self,**parse_kwargs):
        """ Parse the report files.
            produces output of this nature:
            { k-index1  : { 'dft_enrgy':[...], 'qp_energy':[...] },
              k-index2  :{...}
            }
            k-index is the kpoint at which the yambo calculation was
            done.
        """
        # start with check for  failure due to error:
        err = re.compile('^\s+?\[ERROR\]\s+?(.*)$')
        kpoints = re.compile('^  [A-X*]+\sK\s\[([0-9]+)\]\s[:](?:\s+)?([0-9.E-]+\s+[0-9.E-]+\s+[0-9.E-]+)\s[A-Za-z()\s*.]+[0-9]+[A-Za-z()\s*.]+([0-9.]+)')
        memory = re.compile('^\s+?<([0-9a-z-]+)> ([A-Z0-9]+)[:] \[M  ([0-9.]+) Gb\]? ([a-zA-Z0-9\s.()\[\]]+)?')
        timing = re.compile('\s+?[A-Za-z]+iming\s+?[A-Za-z/\[\]]+[:]\s+?([a-z0-9-]+)[/]([a-z0-9-]+)[/]([a-z0-9-]+)')
        self.memstats.extend([ line for line in self.lines if memory.match(line)])
        for line in self.lines:
            if err.match(line):
                if 'STOP' in err.match(line).groups()[0]:
                    # stop parsing, this is a failed calc.
                    self.errors.append(err.match(line).groups()[0])
                    return
            if timing.match(line):
                self.timing.append(timing.match(line).groups()[0] )
            if kpoints.match(line):
                kindx, kpt, wgt = kpoints.match(line).groups()
                self.kpoints[str(int(kindx))] =  [ float(i.strip()) for i in kpt.split()]

        full_lines = '\n'.join(self.lines)
        qp_regx = re.compile('(^\s+?QP\s\[eV\]\s@\sK\s\[\d+\][a-z0-9E:()\s.-]+)(.*?)(?=^$)',re.M|re.DOTALL)
        kp_regex = re.compile('^\s+?QP\s\[eV\]\s@\sK\s\[(\d+)\][a-z0-9E:()\s.-]+$')
        spliter = re.compile('^(B[=]\d+\sEo[=]\s+?[E0-9.-]+\sE[=]\s+?[E0-9.-]+\sE[-]Eo[=]\s+?[E0-9.-]+\sRe[(]Z[)][=]\s+?[E0-9.-]+\sIm[(]Z[)][=]\s?[E0-9.-]+\snlXC[=]\s+?[E0-9.-]+\slXC[=]\s+?[E0-9.-]+\sSo[=]\s+?[E0-9.-]+)')
        extract = re.compile('B[=](\d+)\sEo[=](?:\s+)?([E0-9.-]+)\sE[=](?:\s+)?([E0-9.-]+)\sE[-]Eo[=](?:\s+)?([E0-9.-]+)\sRe[(]Z[)][=](?:\s+)?([E0-9.-]+)\sIm[(]Z[)][=](?:\s+)?[E0-9.-]+\snlXC[=](?:\s+)?([E0-9.-]+)\slXC[=](?:\s+)?([E0-9.-]+)\sSo[=](?:\s+)?([E0-9.-]+)')
        qp_lines = qp_regx.findall(full_lines)
        qp_results ={}
        for each in qp_lines: # first group of qp data, shares k-point index
            kp_index = None
            kp_results={'bindex':[],'dft_energy':[],'qp_energy':[],'qp_correction':[],
                        'z_factor':[],'non_local_xc':[],'local_xc':[],'selfenergy_c':[]}
            for line in each: # different band indexes =B
                if kp_regex.match(line):
                    kp_index = str(int(kp_regex.match(line).groups()[0]))
                else: #  data line B=x Eo = y ..
                    data_lines = [ i for i in spliter.split(line) if i.strip()]
                    for qp_data in data_lines:
                        bindex, dft_energy, qp_energy, qp_correction, z_factor, \
                        non_local_xc, local_xc, selfenergy_c = [float (i) for i in extract.match(qp_data).groups()]
                        kp_results['bindex'].append(bindex)
                        kp_results['dft_energy'].append(dft_energy)
                        kp_results['qp_energy'].append(qp_energy)
                        kp_results['qp_correction'].append(qp_correction)
                        kp_results['z_factor'].append(z_factor)
                        kp_results['non_local_xc'].append(non_local_xc)
                        kp_results['local_xc'].append(local_xc)
                        kp_results['selfenergy_c'].append(selfenergy_c)
            qp_results[kp_index] = kp_results
        self.data = qp_results

    def get_type(self):
        """ Get the type of file
        """
        return self.type

    def has_errors(self):
        #check if the list is empty
        return not not self.errors

    def get_errors(self):
        """ Check if this is a report file and if it contains errors
        """
        if self.type == 'report':
            return self.errors
        return False

    def get_data(self):
        """ Get the data from this file as a dictionary
        """
        pass

    def parse_log(self,**parse_kwargs):
        """ Get ERRORS and WARNINGS from  l-*  file, useful for debugging
        """
        warning = re.compile('^\s+?<([0-9a-z-]+)> ([A-Z0-9]+)[:] \[(WARNING)\]? ([a-zA-Z0-9\s.()\[\]]+)?')
        error = re.compile('^\s+?<([0-9a-z-]+)> ([A-Z0-9]+)[:] \[(ERROR)\]? ([a-zA-Z0-9\s.()\[\]]+)?')
        self.warnings.extend([ line for line in self.lines if warning.match(line)])
        self.errors.extend([ line for line in self.lines if error.match(line)])


    def __bool__(self):
        if self.type == 'unknown':
            return False
        else:
            return True
    __nonzero__=__bool__

    def __str__(self):
        return "type: %9s   file: %s/%s"%(self.type, self.folder, self.filename)




import matplotlib.pyplot as plt
from yambopy import *

y = YamboOut('bse',save_folder='bse/SAVE')

energy = y.files['o-yambo.eps_q1_diago_bse']['E/ev[1]']
im_eps = y.files['o-yambo.eps_q1_diago_bse']['EPS-Im[2]']

plt.plot(energy,im_eps)
#plt.savefig('bse.png')
plt.show()



#
# License-Identifier: GPL
#
# Copyright (C) 2024 The Yambo Team
#
# Authors: HPC
#
# This file is part of the yambopy project
#
"""import a special json encoder to do slightly readable files"""
import json
import re
import numpy as np

class YambopyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (np.ndarray,np.number)):
            if np.iscomplexobj(obj):
                return [obj.real, obj.imag]
            else:
                return obj.tolist()
        return(json.JSONEncoder.default(self, obj))

def JsonLoaders(string):
    """ load json from string """
    return json.loads(filename)

def JsonLoader(filename):
    """ load json from file """
    with open(filename,'r') as f:
        data = json.load(f)
    return data

def JsonDumpers(data):
    """ dump dicitonary as string """
    s = json.dumps(data,cls=YambopyEncoder,indent=2)
    return re.sub(r'\s+([0-9\]\-])','\\1',s)

def JsonDumper(data,filename):
    """ dump dictionary as file """
    with open(filename,'w') as f:
        f.write(JsonDumpers(data))



#!/usr/bin/env python
#TODO: delete/move any other scripts in this directory
from yambopy import *
from yambocommandline.commands import *
import argparse
import sys
import importlib.metadata

class Cmd():
    """
    Define some generic functions for a command
    """
    def info(self):
        """
        display the available commands
        """
        print('yambopy v{}'.format(importlib.metadata.version('yambopy')))
        print('Available commands are:\n')
        for cmd,c in list(self._commands.items()):
            print("%15s -> %s"%(cmd, c.__doc__.split('\n')[1]))

    def cite(self):
        """
        Display the BibTeX citation string
        """
        version=importlib.metadata.version('yambopy')
        print("@misc{yambopy,")
        print("      author = {Paleari, Fulvio and Molina-Sánchez, Alejandro and Nalabothula, Muralidhar and Reho, Riccardo and Bonacci, Miki and Castelo, José M. and Cervantes-Villanueva, Jorge and Pionteck, Mike and Silvetti, Martino and Attaccalite, Claudio and Pereira Coutada Miranda, Henrique},")
        print("      title = {Yambopy},")
        print("      month = mar,")
        print("      year = 2025,")
        print("      publisher = {Zenodo},")
        print("      version = {%s},"%version)
        print("      doi = {10.5281/zenodo.15012962},")
        print("      url = {https://doi.org/10.5281/zenodo.15012962},}")
 
    def run(self,cmds,args):
        """
        generic run command
        cmds is a dictionary that maps the command to the function to run
        """
        cmd = args[0]
        if cmd in list(cmds.keys()):
            cmds[cmd](args[1:])

class PlotExcitons(Cmd):
    """
    Plot excitons calculation.
        
        possible arguments are:
        
        Arguments:
        filename -> json file containing the absorption spectra. Default: 'absorptionspectra.json' 
        -s       -> Size of the materis in the plot
    """
    def __init__(self,args):
        import matplotlib

        #check for args
        if len(args) < 1:
            print((self.__doc__))
            exit(0)

        parser = argparse.ArgumentParser(description='Study convergence on BS calculations using ypp calls.')
        pa = parser.add_argument
        pa('filename',     help='json file containing the absorption spectra. Default: \'absorptionspectra.json\'' )
        pa('-s','--size',  help='Size of the markers in the plot', default=20, type=int)
        args = parser.parse_args(args)

        if os.path.isfile(args.filename):
            #create plot
            recipes.plot_excitons(args.filename,size=args.size) 
        else:
            print('file %s is invalid'%filename)

class PlotEm1sCmd(Cmd):
    """
    Plot em1s calculation.

        possible arguments are:
        
        Arguments:
        folders        -> Folders containing the ndb.ems1 files
        -w, --write    -> Write data file in a text file
        -p, --plot     -> Save a file with the plot
        -v, --verbose  -> Print which files are not folder
        --fontsize     -> Choose the font size of the plot
    """
    def __init__(self,args):
        import matplotlib
        import matplotlib.pyplot as plt        
        from glob import glob

        #check for args
        if len(args) < 1:
            print((self.__doc__))
            exit(0)

        parser = argparse.ArgumentParser(description='Plot em1s calculation.')
        pa = parser.add_argument
        pa('folders',        nargs='+', help='json file containing the absorption spectra. Default: \'absorptionspectra.json\'' )
        pa('-w','--write',   help='Write data file in a text file', default='em1s.dat', type=str)
        pa('-p','--plot',    help='Save a file with the plot', default='em1s.pdf', type=str)
        pa('-v','--verbose', help='Print which files are not folder', action='store_true')
        pa('--fontsize',     help='Choose the font size of the plot', default=10, type=int)
        args = parser.parse_args(args)
        folders = args.folders
        #print(glob('bse_cutoff/*/*'))
        #print(folders)
        #exit() 
        #create plot
        fig = plt.figure(figsize=(6,5))
        matplotlib.rcParams.update({'font.size': args.fontsize})
        ax = plt.gca()

        epsilons = []
        for folder in folders: 
            if os.path.isdir(folder) and os.path.isfile("%s/ndb.em1s"%folder):
                if not os.path.isfile("%s/ns.db1"%folder):
                    if not os.path.isfile("%s/../SAVE/ns.db1"%folder):
                        if args.verbose: print("SAVE folder not found")
                    else:
                        os.system('cp %s/../SAVE/ns.db1 %s/'%(folder,folder))
                print(folder)
                ys = YamboStaticScreeningDB(save=folder,em1s=folder)
                #plot epsilon^-1_{00} = [(1+vX)]_{00}
                ys.plot_epsm1(ax,marker='o',markersize=2,label=folder)
                #get epsilon^{-1}_{00} = [1+vX]_{00}
                x,vX = ys._getvxq()
                epsilons.append([folder,x,1+vX])
            else:
                if args.verbose:
                    print("path %s is not a folder"%folder)

        if args.write:
            #write a text file with the data
            f=open(args.write,'w')
            f.write('# Real and imaginary part of \\epsilon^{-1}_{00}(\\omega=0,q) = [1+vX]_{00} as a funciton of |q|')
            for folder,x,y in epsilons:
                f.write('#%s\n'%folder)
                for xi,yi in zip(x,y):
                    f.write("%12.8lf %12.8lf %12.8lf\n"%(xi,yi.real,yi.imag))
                f.write('\n\n')
            f.close()

        plt.legend(frameon=False)
        plt.tight_layout()

        #final plot
        if args.write: plt.savefig(args.plot) 
        plt.show()

class AnalyseGWCmd(Cmd):
    """
    Study the convergence of GW calculations by looking at the change in band-gap value.

    The script reads from <folder> all results from <variable> calculations and display them.

    Use the band and k-point options according to the size of your k-grid
    and the location of the band extrema.

        Mandatory arguments are:
            folder   -> Folder containing SAVE and convergence runs.
            var      -> Variable tested (e.g. FFTGvecs)

        Optional variables are:
            -bc, --bandc   (int)  -> Lowest conduction band number
            -kc, --kpointc (int)  -> k-point index for conduction band
            -bv, --bandv   (int)  -> Highest valence band number
            -kv, --kpointv (int)  -> k-point index for valence band
            -np, --nopack  (flag) -> Do not call 'pack_files_in_folder'
            -nt, --notext  (flag) -> Do not print a text file
            -nd, --nodraw  (flag) -> Do not draw (plot) the result
    """

    def __init__(self,args):

        #check for args
        if len(args) <= 1:
            print((self.__doc__))
            exit(0)

        #all the other arguments are passed to the analysegw function

        parser = argparse.ArgumentParser(description='Study GW convergence with regards to the band-gap value.')
        pa = parser.add_argument
        pa('folder'            , help='Folder containing SAVE and convergence runs.')
        pa('variable'          , help='Variable tested (e.g. FFTGvecs)' )
        pa('-bc','--bandc'     , help='Lowest conduction band number'    , default=53, type=int)
        pa('-kc','--kpointc'   , help='K-point index for conduction band', default=19, type=int)
        pa('-bv','--bandv'     , help='Highest valence band number'      , default=52, type=int)
        pa('-kv','--kpointv'   , help='K-point index for valence band'   , default=1 , type=int)
        pa('-np','--no-pack'   , help='Skip the packing of output files'        , dest='pack', action='store_false')
        pa('-nt','--no-text'   , help='Skip the writing of the analysis result' , dest='text', action='store_false')
        pa('-nd','--no-draw'   , help='Skip the plotting of the analysis result', dest='draw', action='store_false')
        pa('-v', '--verbose'   , action='store_false')
        parser.set_defaults(pack=True,text=True,draw=True)
        args = parser.parse_args(args)

        folder = args.folder ; var     = args.variable
        bandc  = args.bandc  ; kpointc = args.kpointc
        bandv  = args.bandv  ; kpointv = args.kpointv
        pack   = args.pack   ; text    = args.text
        draw   = args.draw   ; verbose = args.verbose
        
        #call analyse_gw from recipes.py
        recipes.analyse_gw(folder,var,bandc,kpointc,bandv,kpointv,pack,text,draw,verbose)

    def info(self):
        """
        display help to use this command
        """
        print(self.__doc__)

class AnalyseBSECmd(Cmd):
    """
    Study the convergence of BSE calculations.

    Using ypp, you can study the convergence of BSE calculations in 2 ways:
      Create a .png of all absorption spectra relevant to the variable you study
      Look at the eigenvalues of the first n "bright" excitons (given a threshold intensity)

    The script reads from <folder> all results from <variable> calculations for processing.
    The resulting pictures and data files are saved in the ./analyse_bse/ folder.

    Mandatory arguments are:
        folder   -> Folder containing SAVE and convergence runs.
        var      -> Variable tested (e.g. FFTGvecs)

    Optional arguments are:
        -ne,--numbexc  (int)   -> Number of excitons to read beyond threshold (default=2)
        -ie,--intexc   (float) -> Minimum intensity for excitons to be considered bright (default=0.05)
        -de,--degenexc (float) -> Energy threshold under which different peaks are merged (eV) (default=0.01)
        -me,--maxexc   (float) -> Energy threshold after which excitons are not read anymore (eV) (default=8.0)
        -np,--nopack   (flag)  -> Skips packing o- files into .json files
        -nt,--notext   (flag)  -> Skips writing the .dat file
        -nd,--nodraw   (flag)  -> Skips drawing (plotting) the abs spectra
    """
    def __init__(self,args):
        #check for args
        if len(args) < 2:
            print((self.__doc__))
            exit(0)

        parser = argparse.ArgumentParser(description='Study convergence on BS calculations using ypp calls.')
        pa = parser.add_argument
        pa('folder',           help='Folder containing SAVE and convergence runs.' )
        pa('variable',         help='Variable tested (e.g. FFTGvecs)' )
        pa('-ne','--numbexc',  help='Number of excitons to read beyond threshold', default=2,type=int)
        pa('-ie','--intexc',   help='Minimum intensity for excitons to be considered bright', default=0.05,type=float)
        pa('-de','--degenexc', help='Energy threshold under which different peaks are merged (eV)', default=0.01,type=float)
        pa('-me','--maxexc',   help='Energy threshold after which excitons are not read anymore (eV)', default=8.0,type=float)
        pa('-nt','--notext',   help='Skips writing the .dat file', action='store_false')
        pa('-nd','--nodraw',   help='Skips drawing (plotting) the abs spectra', action='store_false')
        args = parser.parse_args(args)

        folder    = args.folder
        var       = args.variable
        exc_n     = args.numbexc
        exc_int   = args.intexc
        exc_degen = args.degenexc
        exc_max_E = args.maxexc
        text      = args.notext
        draw      = args.nodraw

        #all the other arguments are passed to the analyse bse function
        recipes.analyse_bse( folder, var, exc_n, exc_int, exc_degen, exc_max_E, text=text, draw=draw )

class MergeQPCmd(Cmd):
    """
    Merge QP databases.
    
        possible arguments are:

           <QP files>    -> list of QP files produced by yambo
        -o <output file> -> output file where to save the merged db
    """

    def __init__(self,args):
        """ 
        possible arguments are:
        """ 
        #check for args
        if len(args) <= 1:
            print((self.__doc__))
            exit(0)
        
        #all the other arguments are passed to the merge_qp function
        parser = argparse.ArgumentParser(description='Join different NetCDF quasi-particle databases')
        parser.add_argument('files', nargs='+', type=argparse.FileType('r'))
        parser.add_argument('-o','--output',                       help='Output filename', default='ndb_out.QP')
        args = parser.parse_args(args)

        output  = args.output
        files   = args.files
        
        #call merge_qp from recipes.py
        recipes.merge_qp(output,files)

    def info(self):
        """
        display help to use this command 
        """
        print(self.__doc__)


class AddQPCmd(Cmd):
    """
    Add corrections from QP databases.
    This function reads the QP correction from Yambo databases and add them together. Pass a list of files (separated by a space) after each flag to get the total correction you need.
    The k-grid must be the same, the number of bands can be different. There will always be at least the LDA value. The Z factor is taken as one.

    Optional arguments are:

        -a,  --add       -> Add the real-part of a QP correction
        -s,  --subtract -> Substract the real-part of a QP correction
        -ai, --addimg    -> Add the imaginary-part of a QP correction
        -o,  --output    -> Output filename
        -v,  --verbose   -> Increased verbosity

    """

    def __init__(self,args):
        #check for args
        if len(args) <= 1:
            print((self.__doc__))
            exit(0)

        parser = argparse.ArgumentParser(description='Join different NetCDF quasi-particle databases.')
        parser.add_argument('-a', '--add',       nargs='+', type=argparse.FileType('r'), help="Add the real part correction E-Eo to the final db",default=[])
        parser.add_argument('-s', '--subtract', nargs='+', type=argparse.FileType('r'), help="Subtract the real part correction E-Eo part to the final db", default=[])
        parser.add_argument('-ai','--addimg',    nargs='+', type=argparse.FileType('r'), help="Add the imaginary part Im(E) to the final db",default=[])
        parser.add_argument('-o', '--output',  default='ndb_out.QP', help='Output filename')
        parser.add_argument('-v', '--verbose', action="store_true",  help='Verbose mode')
        args = parser.parse_args(args)


        output  = args.output
        add     = args.add
        subtract = args.subtract
        addimg  = args.addimg
        verbose = args.verbose


        #call add_qp from recipes.py
        recipes.add_qp(output,add,subtract,addimg,verbose)

    def info(self):
        """
        display help
        """
        print(self.__doc__)

class SaveCmd(Cmd):
    """
    Produce a SAVE folder.
    
    Arguments are:
        -nscf, --nscf_dir  -> Path to nscf save folder
        -y, --yambo_dir    -> <Optional> Path to yambo executables
    """
    def __init__(self,args):
        #check for args
        if len(args) < 1:
            print((self.__doc__))
            exit(0)
            
        parser = argparse.ArgumentParser(description='Generate SAVE folder including gkkp databases')
        parser.add_argument('-nscf','--nscf_dir', type=str,help='<Required> Path to nscf save folder', required=True)
        parser.add_argument('-y','--yambo_dir', default="", type=str,help='<Optional> Path to yambo executables')
        args = parser.parse_args(args)
        
        nscf_dir  = args.nscf_dir
        yambo_dir = args.yambo_dir
        
        database = './'
        scheduler = Scheduler.factory

        #call generate_save
        generate_save.generate_save(database,nscf_dir,yambo_dir,scheduler)

class UpdtSrlNmbrCmd(Cmd):
    """
    Script to update serial numbers of yambo ndb.* databases in order to import them to new calculations.

    Inputs:
    1. -new, --new_serial='path/to/folder/with/new/dbs' [e.g., the new SAVE]
    2. -old, --old_serial='path/to/folder/with/old/dbs' [e.g., an old ndb.em1s]

    This script will prompt the user to go through with updating the dbs.
    """
    def __init__(self,args):
        #check for args
        if len(args) < 2:
            print((self.__doc__))
            exit(0)

        parser = argparse.ArgumentParser(description='Updated serial numbers in yambo databases')
        parser.add_argument('-new','--new_serial', type=str, default="./SAVE", help='<Optional> Path to folder with the newer databases (Default is ./SAVE)')
        parser.add_argument('-old','--old_serial', type=str,help='<Required> Path to folder with the older databases',required=True)
        args = parser.parse_args(args)

        new_dir = args.new_serial
        old_dir = args.old_serial

        #check dbs
        update_serial.prompt_user(new_dir,old_dir)

class PlotBndStrCmd(Cmd):
    """
    Script to produce band structure data and visualization from QE.
    
    It reads data-file-schema.xml found in the save folders of newer pw versions.
    
    It requires a yaml input file (see below)

    Arguments are:
        -i, --band_input   -> Path to yaml input file
        -o, --output_name  -> Name of output file <optional>
        -e, --erange       -> Energy window for plot <optional>
        -s, --show         -> Show plot at runtime <optional>

    - YAML INPUT FILE 
    This is a yaml file in the following format, to be copy-pasted and edited:
    
    :: yaml
    
        ---
         save_dir: "PREFIX.save"
         KPTs: [[0.,0.,0.],[0.,0.5,0.],[0.5,0.5,0.],[0.5,0.,0.]]
         KPTs_labels: [G, Y, S, X]
         shift_Delta_c_v: [0.,1.,1.]
    
    :: 
    
    The input parameters are:
    - save_dir: path to folder containing 'data-file-schema.xml' (in principle the QE save folder)
    - KPTs: band circuit in reduced coordinates
    - KPTs_labels: labels for the band circuit points
    - shift_Delta_c_v: k-dependent scissor shift as a list of three values (gap shift, cond. stretch, val. stretch)
    """
    def __init__(self,args):

        #check for args
        if len(args) < 2:
            print((self.__doc__))
            exit(0)

        parser = argparse.ArgumentParser(description='Generate band plot')
        parser.add_argument('-i','--band_input', type=str,  help='<Required> Path to input file',      required=True)
        parser.add_argument('-o','--output_name', type=str,  help='<Optional> Name of output file',    required=False, default=None)
        parser.add_argument('-e','--erange',      type=float,help='<Optional> Energy window for plot', required=False, default=4.)
        parser.add_argument('-s','--show',        type=bool, help='<Optional> Show plot at runtime',   required=False, default=False)
        args = parser.parse_args(args)

        inp = args.band_input

        # Read user input file
        input_params = generate_bands.read_input(inp)

        # Read data from .xml file
        output_data = generate_bands.get_data_from_xml(input_params)

        # Plot setup
        data_to_plot = generate_bands.setup_BZ_points(input_params,output_data)

        # Produce the plot
        generate_bands.launch_plot(data_to_plot,plt_type='bands',out_name=args.output_name,erange=args.erange,show=args.show)

class GwSubspace(Cmd):
    """
    Script to calculate off-diago corrections of yambo ndb.QP databases in order to plot band structure.

    Inputs:
    1. -d,--fld_diag='path/to/folder/with/diago/dbs' [e.g., the ./diag]
    2. -o,--fld_offdiag='path/to/folder/with/offdiagoold/dbs' [e.g., the ./offdiago]

    This script will prompt the user to go through with updating the dbs.
    """
    def __init__(self,args):
        #check for args
        if len(args) < 2:
            print((self.__doc__))
            exit(0)

        parser = argparse.ArgumentParser(description='Create a new diago dbs after a GW subspace calculation numbers in yambo databases')
        parser.add_argument('-d','--fld_diag', type=str, default="./diago", help='<Required> Path to folder with the ndb.QP diago-database   (Default is ./diago)',required=True)
        parser.add_argument('-o','--fld_offdiag', type=str,default ="./offdiago",help='<Required> Path to folder with the ndb.QP offdiagodiago-database   (Default is ./diago)',required=True)
        args = parser.parse_args(args)

        fld_diag = args.fld_diag
        fld_offdiag = args.fld_offdiag
        gw_subspace.create_newdb(fld_diag,fld_offdiag)
class ConvertRLtoRyCmd(Cmd):
    """
    Script to convert RL number in Ry energy units using ndb.gops.

    Inputs:
    1. --ndb_gops path/to/SAVE/folder [or any folder containing ndb.gops]
    2. --value value to convert with units [e.g., 11 RL or 5 Ry]

    The script will read ndb.gops and find the nearest completed G-shell, then give the
    converted value in Ry (RL) to the one supplied in input.
    """
    def __init__(self,args):

        #check for args
        if len(args) < 4:
            print((self.__doc__))
            exit(0)

        parser = argparse.ArgumentParser(description='Convert RL number in energy units (Ry)')
        parser.add_argument('-gops','--ndb_gops', type=str, default="./SAVE", help='<Optional> Path to folder with ndb.gops (default: ./SAVE)')
        parser.add_argument('-v','--value', type=str,help="<Required> Value to be converted along with units, e.g.: '11 RL' or '5 Ry'",nargs=2,required=True)
        args = parser.parse_args(args)

        ndb_gops = args.ndb_gops
        value    = args.value

        convert_RL_to_Ry.convert(value,ndb_gops)

class ConvertLELPHCtoYAMBO(Cmd):
    """
    Calculate gauge-invariant electron-phonon matrix elements with LetzElPhC and convert them into Yambo format.
    
    - Usage:
    
    >> yambopy l2y -ph phinp -b b1 b2 -par nq nk [--kernel kernel] [--lelphc lelphc] [--no_lelphc_dbs] [--no_gkkp]
    
    - Input parameters:
        -ph                   : path to ph.x input file, e.g. dvscf/ph.in
        -b                    : initial and final band indices (counting from 1, both values included)
        -par [OPT]            : MPI pools for q and k (needs mpirun)
        --kernel [OPT]        : e-ph kernel type, default 'dfpt'
        --lelphc [OPT]        : path to lelphc executable, default 'lelphc', code will prompt
        --no_lelphc_dbs [OPT] : will remove LetzElPhC input and outputs
        --no_gkkp [OPT]       : do not generate the gkkp databases for Yambo/Lumen

    - Prerequisites:

    * ph.x phonon calculation must be complete, e.g. the phinp folder should contain:
        - ph.x input file
        - pw.x (scf) save directory
        - [prefix].dyn files
        - _ph* directories
    * Yambo SAVE directory must be present. We run in the directory where the SAVE is.
    * LetzElPhC must be installed
    * mpirun must be linked for parallel runs
    """
    def __init__(self,args):

        #check for args
        if len(args) < 5:
            print((self.__doc__))
            exit(0)

        parser = argparse.ArgumentParser(description='Generate electron-phonon coupling databases via LetzElPhC')
        parser.add_argument('-ph','--ph_inp_path', type=str, help='<Required> Path to ph.x (dvscf) input file',required=True)
        parser.add_argument('-b','--bands',nargs=2,type=str,help="<Required> First and last band (counting from 1), e.g. 'b_i b_f'",required=True)
        parser.add_argument('-k','--kernel', type=str, default='dfpt',help="<Optional> Electron-phonon kernel type, e.g. 'dfpt', 'bare', ... (default 'dfpt')")
        parser.add_argument('-par','--pools',nargs=2,type=str, default=[1,1], help="<Optional> MPI tasks as 'nqpools nkpools' (default serial)")
        parser.add_argument('-lelphc','--lelphc',type=str,default='lelphc',help="<Optional> Path to lelphc executable (default assumed in Path, otherwise prompted)")
        parser.add_argument('-nl','--no_lelphc_dbs', action="store_true", help="Remove lelphc databases (False if not given)")
        parser.add_argument('-ng','--no_gkkp', action="store_true", help="Do not generate gkkp dbs")

        args = parser.parse_args(args)

        phinp         = args.ph_inp_path
        bands         = args.bands
        kernel        = args.kernel
        pools         = args.pools
        lelphc        = args.lelphc
        no_lelphc_dbs = args.no_lelphc_dbs
        no_gkkp       = args.no_gkkp

        # Check inputs
        lelphc,ph_path,inp_ph,inp_lelphc,inp_name = \
        lelph_interface.checks(phinp,lelphc,bands,kernel,pools)

        # run preprocessing
        lelph_interface.run_preprocessing(lelphc,ph_path,inp_ph)

        # run el-ph calculation and rotation
        lelph_interface.run_elph(lelphc,inp_lelphc,inp_name)

        # load database and convert to yambo format
        if not no_gkkp:
           lelph_interface.letzelph_to_yambo()

        # clean
        lelph_interface.clean_lelphc(no_lelphc_dbs,inp_name,ph_path)

class BSEKernelSizeCmd(Cmd):
    """
    Script to calculate the expected BSE kernel size, in GB, to be loaded by each MPI task

    Size_kernel_GB = (nk*nv*nc*ncpl)**2*np*4*2)/(1024.)**3.

    Usage:

    >> yambopy bsesize -nk nk -nv nv -nc nc [-ndp] [-ncpl]

    :: -nk,--nkpts_bz      = Number of kpoints in the full BZ
    :: -nv,--nvalence,     = Number of valence bands included in the kernel
    :: -nc,--nconduction   = Number of conduction bands included in the kernel
    :: -ndp,--ndoubleprec  = Double precision [Default: False]
    :: -ncpl,--ncoupling   = BSE kernel in coupling mode (i.e., no TDA) [Default: False ]

    """
    def __init__(self,args):

        #check for args
        if len(args) < 5:
            print((self.__doc__))
            exit(0)

        parser = argparse.ArgumentParser(description='Calculate expected size of BSE kernel to be loaded by each MPI task (in GB)')
        parser.add_argument('-nk',   '--nkpts_bz', type=int, help='Number of kpoints in the full BZ',required=True)
        parser.add_argument('-nv',   '--nvalence', type=int, help='Number of valence bands included in the kernel',required=True)
        parser.add_argument('-nc','--nconduction', type=int, help='Number of conduction bands included in the kernel',required=True)
        parser.add_argument('-ndp', '--ndoubleprec', help='double precision',action='store_false')
        parser.add_argument('-ncpl','--ncoupling', help='BSE kernel in coupling mode (i.e., no TDA)',action='store_false')

        args = parser.parse_args(args)

        nk   = args.nkpts_bz
        nv   = args.nvalence
        nc   = args.nconduction
        np   = 1 if args.ndoubleprec else 2
        ncpl = 1 if args.ncoupling else 2

        get_BSE_kernel_size.get_BSE_kernel_size(nk,nv,nc,np,ncpl)

class Exc_sortCmd(Cmd):
    """
    Script to sort excitonic states according to energies and intensities. [-h for help]
    """
    def __init__(self,args):
        from yambocommandline.commands import exc_sort
        exc_sort.run_exc_sort(args)

class Exc_wf_plotCmd(Cmd):
    """
    Real space exciton wf for either fixed hole/electron. [-h for help]
    """
    def __init__(self,args):
         from yambocommandline.commands import plot_exc_wf_real_space
         plot_exc_wf_real_space.run_plot_exc_wf_real_space(args)


class Exc_irrepCmd(Cmd):
    """
    Script to get the irreducible representation labels for excitonic states. [-h for help]
    """
    def __init__(self,args):
        from yambocommandline.commands import run_exc_irreps
        run_exc_irreps.run_exc_irrep(args)

class GkkpCmd(Cmd):
    """
    [LEGACY] Produce a SAVE folder including elph_gkkp databases.
    
    Arguments are:
        -nscf, --nscf_dir  -> <Optional> Path to nscf save folder
        -elph, --elph_dir  -> Path to elph_dir folder
        -y, --yambo_dir    -> <Optional> Path to yambo executables
        -e, --expand       -> <Optional> Expand gkkp databases
    """
    def __init__(self,args):
        #check for args
        if len(args) < 2:
            print((self.__doc__))
            exit(0)
            
        parser = argparse.ArgumentParser(description='Generate SAVE folder including gkkp databases')
        parser.add_argument('-nscf','--nscf_dir', type=str, default="", help='<Optional> Path to nscf save folder')
        parser.add_argument('-elph','--elph_dir', type=str,help='<Required> Path to elph_dir folder',required=True)
        parser.add_argument('-y','--yambo_dir', default="", type=str,help='<Optional> Path to yambo executables')
        parser.add_argument('-e','--expand', action="store_true", help="Expand GKKP")
        args = parser.parse_args(args)
        
        nscf_dir  = args.nscf_dir
        elph_dir  = args.elph_dir
        yambo_dir = args.yambo_dir
        expand    = args.expand
        
        database = './'
        scheduler = Scheduler.factory

        #call gkkp
        gkkp.generate_gkkp(database,nscf_dir,elph_dir,yambo_dir,expand,scheduler)
        

class GetPHqInputCmd(Cmd):        
    """
    [LEGACY] Script to update the explicit list of q-points in a ph input file (ldisp=.false., qplot=.true.).

    - Reads the output of the scf calculation and the ph input.

    Usage:
    :: -pw,--pwout='path/to/pw/output/file
    :: -ph,--phin='path/to/ph/input/file'
    """
    def __init__(self,args):

        #check for args
        if len(args) < 4:
            print((self.__doc__))
            exit(0)

        parser = argparse.ArgumentParser(description='Append explicit qpoints to ph input')
        parser.add_argument('-pw','--pwout', type=str, help='Path to pw (scf) output file',required=True)
        parser.add_argument('-ph','--phin', type=str,help='Path to ph (dvscf or elph) input file',required=True)
        args = parser.parse_args(args)

        pwout = args.pwout
        phin  = args.phin

        get_phq_input.get_phq_input(pwout,phin)


class YambopyCmd(Cmd):
    """
    class to implement commands for yambopy.
    each new command to be added should be implemented as a class inheriting from this one
    """
    _commands = {'plotem1s':     PlotEm1sCmd,
                 'analysebse':   AnalyseBSECmd,
                 'analysegw':    AnalyseGWCmd,
                 'plotexcitons': PlotExcitons,
                 'addqp':        AddQPCmd,
                 'mergeqp':      MergeQPCmd,
                 'save':         SaveCmd,
                 'bands':        PlotBndStrCmd,
                 'serial':       UpdtSrlNmbrCmd,
                 'gwsubspace':   GwSubspace,
                 'convert':      ConvertRLtoRyCmd,
                 'bsesize':      BSEKernelSizeCmd,
                 'l2y':          ConvertLELPHCtoYAMBO,
                 'exc-irrep':    Exc_irrepCmd,
                 'exc-wf':       Exc_wf_plotCmd,
                 'exc-sort':     Exc_sortCmd,
                 'gkkp':         GkkpCmd,
                 'phinp':        GetPHqInputCmd,
                 }

    def __init__(self,*args):
        """
        parse the command from the command line and initialize the class responsible
        for handling such command
        """
        
        #check for args
        if len(args) <= 1:
            self.info()
            exit(0)

        #check for how to cite
        if args[1]=='cite':
            self.cite()
            exit()

        #start call graph     
        if args[1] in self._commands:
            cmdclass = self._commands[args[1]]
            self.cmd = cmdclass(args[2:]) 
        else:
            self.info()  
            print() 
            print("Command %s is not known to yambopy"%args[1])
 
#parse options
#def run_script(): return YambopyCmd(*sys.argv)
ycmd = YambopyCmd(*sys.argv)
exit()



# Copyright (C) 2018 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yambopy
#
import unittest
import os
from yambopy.dbs.dipolesdb import YamboDipolesDB
from yambopy.dbs.latticedb import YamboLatticeDB
from yambopy.dbs.electronsdb import YamboElectronsDB

test_path = os.path.join(os.path.dirname(__file__),'..','..','data','refs','ip','SAVE')

class TestYamboDipolesDB(unittest.TestCase):

    def test_yambodipolesdb(self):

        # read lattice
        lat = YamboLatticeDB.from_db_file(os.path.join(test_path,'ns.db1'))

        #read electrons
        electrons = YamboElectronsDB(lat,save=test_path) 
        print(electrons)

        #read dipoles
        # P  -> the velocity matrix elements
        # iR -> from the velocity matrix elements using r = [H,r]
        dipoles = YamboDipolesDB.from_db_file(lat,filename=f'{test_path}/ndb.dipoles',dip_type='P')

        #calculate epsilon
        dipoles.ip_eps2(electrons)

        print(dipoles)

if __name__ == '__main__':
    unittest.main()



#
# Convergence GW on hexagonal BN
# Alejandro Molina-Sanchez & Henrique P. C. Miranda
#
from __future__ import print_function
import sys
from yambopy     import *
from qepy        import *
from schedulerpy import *
import argparse
import matplotlib.pyplot as plt

yambo = 'yambo'
p2y = 'p2y'
prefix = 'bn'

bash = Scheduler.factory

def create_save():
    #check if the nscf cycle is present
    if os.path.isdir('nscf/%s.save'%prefix):
        print('nscf calculation found!')
    else:
        print('nscf calculation not found!')
        exit() 

    #check if the SAVE folder is present
    if not os.path.isdir('database'):
        print('preparing yambo database')
        shell = bash() 
        shell.add_command('mkdir -p database')
        shell.add_command('cd nscf/%s.save; %s; %s'%(prefix, p2y, yambo))
        shell.add_command('mv SAVE  ../../database/')
        shell.run()
        shell.clean()

def gw_convergence():
    #create the folder to run the calculation
    if not os.path.isdir('gw_conv'):
        shell = bash() 
        shell.add_command('mkdir -p gw_conv')
        shell.add_command('cp -r database/SAVE gw_conv/')
        shell.run()
        shell.clean()

    y = YamboIn.from_runlevel('%s -p p -g n -V all'%yambo,folder='gw_conv')
    k_i = y['QPkrange'][0][0]         # Read the first k-point in the uniform k-grid
    k_f = y['QPkrange'][0][1]         # Read the last k-point in the uniform k-grid
    print('The K-point is at the k-index %d' % k_f)

    y['BndsRnXp'] = [[1,10],'']             # Screening. Number of bands
    y['NGsBlkXp'] = [0,'Ry']                # Cutoff Screening
    y['GbndRnge'] = [[1,10],'']             # Self-energy. Number of bands
    y['QPkrange'] = [ [k_i,k_f,4,5], '' ]

    conv = { 'EXXRLvcs': [[10,10,20,30,40,50,60,70,80,90,100],'Ry'],
             'NGsBlkXp': [[0,0,1,2,3,4,5,6,7,8,9,10,11,12], 'Ry'],
             'BndsRnXp': [[[1,10],[1,10],[1,20],[1,30],[1,40],[1,50],[1,60],[1,70]],''] ,
             'GbndRnge': [[[1,10],[1,10],[1,20],[1,30],[1,40],[1,50],[1,60],[1,70]],''] }

    def run(filename):
        """ Function to be called by the optimize function """
        folder = filename.split('.')[0]
        print(filename,folder)
        shell = bash() 
        shell.add_command('cd gw_conv')
        shell.add_command('rm -f *.json %s/o-*'%folder) #cleanup
        shell.add_command('%s -F %s -J %s -C %s 2> %s.log'%(yambo,filename,folder,folder,folder))
        shell.run()
        shell.clean()

    y.optimize(conv,folder='gw_conv',run=run,ref_run=False)

def plot_convergence():
    y = YamboIn.from_runlevel('%s -p p -g n -V all'%yambo,folder='gw_conv')
    k_f = y['QPkrange'][0][1]         # Read the last k-points in the uniform k-grid

    print('Plots of band gap convergence for BN (.png)')
    shell = bash() 
    shell.add_command('yambopy analysegw -bc 5 -kc %s -bv 4 -kv %s gw_conv EXXRLvcs' % (k_f, k_f))
    shell.add_command('yambopy analysegw -bc 5 -kc %s -bv 4 -kv %s gw_conv NGsBlkXp' % (k_f, k_f))
    shell.add_command('yambopy analysegw -bc 5 -kc %s -bv 4 -kv %s gw_conv BndsRnXp' % (k_f, k_f))
    shell.add_command('yambopy analysegw -bc 5 -kc %s -bv 4 -kv %s gw_conv GbndRnge' % (k_f, k_f))
    shell.run()
    shell.clean()

def xi():
    #create the folder to run the calculation
    if not os.path.isdir('gw-xi'):
        shell = bash() 
        shell.add_command('mkdir -p gw-xi')
        shell.add_command('cp -r database/SAVE gw-xi/')
        shell.run()
        shell.clean()

    print ("Running COHSEX in folder 'gw-xi/coh'")
    cohsex = YamboIn.from_runlevel('%s -p c -g n -V all'%yambo,folder='gw-xi')
    cohsex['EXXRLvcs'] = [60,'Ry']       # Self-energy. Exchange
    cohsex['BndsRnXs'] = [1,40]          # Screening. Number of bands
    cohsex['NGsBlkXs'] = [3,'Ry']        # Cutoff Screening
    cohsex['GbndRnge'] = [1,30]          # Self-energy. Number of bands
    cohsex['QPkrange'][0][2:] = [1,6]
    cohsex.write('gw-xi/yambo_cohsex.in')
    shell = bash() 
    shell.add_command('cd gw-xi')
    shell.add_command('rm -f coh.json coh/o-coh*') #cleanup
    shell.add_command('%s -F yambo_cohsex.in -J coh -C coh' % yambo)
    shell.run()
    shell.clean()

    print ("Running PPA in folder 'gw-xi/pp'")
    ppa = YamboIn.from_runlevel('%s -p p -g n -V all'%yambo,folder='gw-xi')
    ppa['EXXRLvcs'] = [60,'Ry']       # Self-energy. Exchange
    ppa['BndsRnXp'] = [1,40]          # Screening. Number of bands
    ppa['NGsBlkXp'] = [3,'Ry']        # Cutoff Screening
    ppa['GbndRnge'] = [1,30]          # Self-energy. Number of bands
    ppa['QPkrange'][0][2:] = [1, 6]       # QP range. All BZ
    ppa.write('gw-xi/yambo_ppa.in')
    shell = bash() 
    shell.add_command('cd gw-xi')
    shell.add_command('rm -f pp.json pp/o-pp*') #cleanup
    shell.add_command('%s -F yambo_ppa.in -J pp -C pp' % yambo)
    shell.run()
    shell.clean()

def plot_xi():

    # Define path in reduced coordinates using Class Path
    npoints = 10
    path = Path([ [[  0.0,  0.0,  0.0],'$\Gamma$'],
                  [[  0.5,  0.0,  0.0],'M'],
                  [[1./3.,1./3.,  0.0],'K'],
                  [[  0.0,  0.0,  0.0],'$\Gamma$']], [int(npoints*2),int(npoints),int(sqrt(5)*npoints)] )

    # Read Lattice information from SAVE
    lat  = YamboLatticeDB.from_db_file(filename='gw-xi/SAVE/ns.db1')
    # Read QP database
    y1   = YamboQPDB.from_db(filename='ndb.QP',folder='gw-xi/coh')
    y2   = YamboQPDB.from_db(filename='ndb.QP',folder='gw-xi/pp')

    # 2. Plot of KS and QP eigenvalues NOT interpolated along the path
    ks_bs_1, qp_bs_1 = y1.get_bs_path(lat,path)
    ks_bs_2, qp_bs_2 = y2.get_bs_path(lat,path)

    # 3. Set Fermi energy at the top of the valence band
    n_top_valence_band = 4

    ks_bs_1.set_fermi(n_top_valence_band)
    ks_bs_2.set_fermi(n_top_valence_band)
    qp_bs_1.set_fermi(n_top_valence_band)
    qp_bs_2.set_fermi(n_top_valence_band)

    fig = plt.figure(figsize=(4,5))
    ax = fig.add_axes( [ 0.20, 0.20, 0.70, 0.70 ])
 
    qp_bs_1.plot_ax(ax,legend=True,color_bands='r',label='QP-GW-COHSEX')
    qp_bs_2.plot_ax(ax,legend=True,color_bands='b',label='QP-GW-PPA')
    plt.ylim((-20,20))
    plt.show()

def dyson_eq():
    #create the folder to run the calculation
    folder_dyson = 'gw-zeros'
    if not os.path.isdir(folder_dyson):
        shell = bash() 
        shell.add_command('mkdir -p %s' % folder_dyson)
        shell.add_command('cp -r database/SAVE %s/' % folder_dyson)
        shell.run()
        shell.clean()

    dyson = YamboIn.from_runlevel('%s -p p -g n -V all'%yambo,folder=folder_dyson)

    dyson['EXXRLvcs'] = [60,'Ry']           # Self-energy. Exchange
    dyson['BndsRnXp'] = [1,40]              # Screening. Number of bands
    dyson['NGsBlkXp'] = [ 3,'Ry']        # Cutoff Screening
    dyson['GbndRnge'] = [1,30]              # Self-energy. Number of bands
    dyson['QPkrange'][0][2:] = [2, 6]

    dyson['DysSolver'] = "n" 
    dyson.write('%s/yambo_newton.in' % folder_dyson)
    dyson['DysSolver'] = "s" 
    dyson.write('%s/yambo_secant.in' % folder_dyson)
    shell = bash() 

    print("calculating with Newton and Secant solver in folder 'gw-zeros/newton' and 'gw-zeros/secant'...")
    shell.add_command('cd %s' % folder_dyson)
    shell.add_command('rm -f *.json newton/o-* secant/o-*') #cleanup
    shell.add_command('%s -F yambo_newton.in -J newton -C newton' % yambo)
    shell.add_command('%s -F yambo_secant.in -J secant -C secant' % yambo)
    shell.run()
    shell.clean()

def plot_dyson():

    # Define path in reduced coordinates using Class Path
    npoints = 10
    path = Path([ [[  0.0,  0.0,  0.0],'$\Gamma$'],
                  [[  0.5,  0.0,  0.0],'M'],
                  [[1./3.,1./3.,  0.0],'K'],
                  [[  0.0,  0.0,  0.0],'$\Gamma$']], [int(npoints*2),int(npoints),int(sqrt(5)*npoints)] )

    # Read Lattice information from SAVE
    lat  = YamboLatticeDB.from_db_file(filename='gw-zeros/SAVE/ns.db1')
    # Read QP database
    y1   = YamboQPDB.from_db(filename='ndb.QP',folder='gw-zeros/newton')
    y2   = YamboQPDB.from_db(filename='ndb.QP',folder='gw-zeros/secant')

    # 2. Plot of KS and QP eigenvalues NOT interpolated along the path
    ks_bs_1, qp_bs_1 = y1.get_bs_path(lat,path)
    ks_bs_2, qp_bs_2 = y2.get_bs_path(lat,path)

    fig = plt.figure(figsize=(4,5))
    ax = fig.add_axes( [ 0.20, 0.20, 0.70, 0.70 ])
 
    qp_bs_1.plot_ax(ax,legend=True,color_bands='r',label='QP-GW-Newton')
    qp_bs_2.plot_ax(ax,legend=True,color_bands='b',label='QP-GW-Secant')

    plt.show()


if __name__ == "__main__":
    #parse options
    parser = argparse.ArgumentParser(description='GW convergence')
    parser.add_argument('-c'  ,'--convergence',  action="store_true", help='Run convergence calculations')
    parser.add_argument('-p'  ,'--plot', action="store_true",         help='Pack into json files and plot the convergence results')
    parser.add_argument('-x'  ,'--xi', action="store_true",           help='GW calculations for several approximations of the Screenning')
    parser.add_argument('-xp' ,'--xp', action="store_true",           help='Plot GW results for COHSEX and PPA')
    parser.add_argument('-z'  ,'--zeros', action="store_true",        help='GW calculations for Newton and Secant Solver')
    parser.add_argument('-zp' ,'--zp', action="store_true",           help='Plot GW results for Newton and Secant Solver')

    args = parser.parse_args()

    if len(sys.argv)==1:
        parser.print_help()
        sys.exit(1)

    create_save()
    if args.convergence:    gw_convergence()
    if args.plot:           plot_convergence()
    if args.xi:             xi()
    if args.xp:             plot_xi()
    if args.zeros:          dyson_eq()
    if args.zp:             plot_dyson()



# Copyright (C) 2018 Henrique Pereira Coutada Miranda
# All rights reserved.
#
# This file is part of yambopy
#
from __future__ import print_function, division
import os
import json
import re
import numpy as np
import matplotlib.pyplot as plt
from yambopy.plot.bandstructure import YambopyBandStructure
from yambopy.kpoints import get_path
from yambopy.tools.duck import isstring
from yambopy.io.inputfile import YamboIn
from yambopy.dbs.latticedb import YamboLatticeDB
from yambopy.plot.plotting import add_fig_kwargs
from yambopy.tools.string import marquee

class YamboAnalyser():
    """
    Class to open multiple ``.json`` files, organize them and plot the data together.
    Used to perform convergence tests
    """
    _colormap = 'rainbow'

    def __init__(self, folder='.'):
        self.folder = folder

        #get all the json files in the folder
        all_filenames  = os.listdir(folder)

        #filter for json files
        json_filenames = [f for f in all_filenames if '.json' in f]

        #read all the json files
        self.jsonfiles = dict()
        for json_filename in json_filenames:
            path = "%s/%s"%(folder, json_filename)
            
            with  open(path,'r') as f:
                self.jsonfiles[json_filename] = json.load(f)

    def get_files_type(self,type,tags=None):
        """
        In all the json files present find the ones of a certain type
        the possible types are:
            report
            netcdf_gw
            netcdf_hf
        """
        json_files = self.jsonfiles
        files = {}

        #iterate over all the json files
        for json_filename in json_files.keys():
            json_file = json_files[json_filename]

            #all the output files in each json file
            for output_filename in json_file["files"]:
                output_file = json_file["files"][output_filename]
                if output_file["type"] == type:
                    filename, extension = os.path.splitext(json_filename)
                    files[filename] =  output_file
       
        #filter files with tags
        if tags:
            if isstring(tags):
                tags = (tags,)
            
            filenames = list(files.keys())
            for filename in filenames:
                if all(tag not in filename for tag in tags):
                    files.pop(filename)

        return files

    def get_colors(self,tags):
        """ 
        !!!!!!
        !!!!!! This function was buggy. Let's check it better
        !!!!!!
        Select the colors according to the number of files to plot
        the files to plot are the ones that have all the tags in their name
        """
        import matplotlib.pyplot as plt
        #count the number of files
        nfiles = 0
        for k in self.jsonfiles.keys():
            for filename in list(self.jsonfiles[k]['files'].keys()):
                nfiles+=all(i in filename for i in tags)

        cmap = plt.get_cmap(self._colormap) #get color map
        colors = [cmap(i) for i in np.linspace(0, 1, nfiles)]

        return colors

    def get_inputfiles_tag(self,tags):
        """
        Get a specific tag from all the .json files from the folders
        You need to write down all the tags that you want to find
        The tags are both for variables in the input file and arguments (runlevels)
        """
        #check if a string was passed and in that case we make it a tuple
        if isinstance(tags,str):
            tags = (tags,)

        inputfiles = self.get_inputfiles()
        inputfiles_tags = dict()

        for k in list(inputfiles.keys()):
            inputfiles_tags[k] = dict()

            # get the current inputfile
            this_inputfile = inputfiles[k]

            #initialize the dictionary
            inputfiles_tags[k] = {'variables':{},'arguments':[]}

            for tag in tags:
                for filename in this_inputfile:
                    
                    # look in variables
                    if tag in list(this_inputfile[filename]['variables'].keys()):
                        variables = this_inputfile[filename]['variables'][tag]
                        inputfiles_tags[k]['variables'][tag] = variables

                    #look in arguments
                    if tag in this_inputfile[filename]['arguments']:
                        inputfiles_tags[k]['arguments'].append(tag)

        return inputfiles_tags

    def get_bands(self,tags=None,path_kpoints=None,type_calc=('ks','gw')):
        """
        Get the gw bands from a gw calculation from a filename

        Arguments:
            json_filename: the name of the json file
            output_filename: the name of the output filename that is in the json file
            type_calc:
               We read from the netcdf file:
               Eo : LDA
               E  : GW
               E-Eo : GW corrections
        """
        #get files that have a gw calculation in them
        gw_files = self.get_files_type('netcdf_gw',tags)

        #check if the dimensions of the files are consistent
        #TODO

        #create bandstructure class
        ks_bandstructure, qp_bandstructure = None, None

        # add bandstructures of all the files
        for filename, content in gw_files.items():
            e0,e0imag = content['Eo']
            e,linewidths = content['E']
            ec,linewidths = content['E-Eo']
            kpoints = content['Kpoint']

            #TODO move this section to YamboFileGW class
            #begin section            

            #get dimensions 
            band_index = np.array(content['Band'],dtype=int)
            band_min, band_max = min(band_index), max(band_index)
            nbands = band_max-band_min+1
           
            kpoint_index = np.array(content['Kpoint_index'],dtype=int)
            kpoint_min, kpoint_max = min(kpoint_index), max(kpoint_index)
            nkpoints = kpoint_max-kpoint_min+1

            #get arrays of bands and kpoints
            bands_e0 = np.zeros([nkpoints,nbands])
            bands_e  = np.zeros([nkpoints,nbands])
            for ei,e0i,ki,ni in zip(e,e0,kpoint_index,band_index):
                nkpoint = ki-kpoint_min
                nband = ni-band_min
                bands_e0[nkpoint,nband] = e0i
                bands_e[nkpoint,nband] = ei
            #end section

            if path_kpoints:
                #get data from json file
                jsonfile = list(self.jsonfiles.values())[0]
                lat = YamboLatticeDB.from_dict(jsonfile['lattice'])
                kpoints, bands_indexes, path_car = get_path(lat.car_kpoints,lat.rlat,None,path_kpoints)  
                bands_e0 = bands_e0[bands_indexes]
                bands_e  = bands_e[bands_indexes] 

            #add bands
            if 'ks' in type_calc: 
                ks_bandstructure = YambopyBandStructure(bands_e0,kpoints,label=filename)
            if 'gw' in type_calc:
                qp_bandstructure = YambopyBandStructure(bands_e, kpoints,label=filename)

        return ks_bandstructure, qp_bandstructure

    @add_fig_kwargs
    def plot_ks(self,path=None,tags=None):
        """
        Use this function to plot the kohn sham energies from a GW calculation
        """
        #get bands from these files
        ks_bands = self.get_bands(tags=tags,path=path,type_calc=('ks'))[0]
        
        #plot the bands
        return ks_bands.plot(show=False)

    @add_fig_kwargs
    def plot_gw(self,path_kpoints=None,tags=None,**kwargs):
        """
        Use this function to plot the quasiparticle energies from a GW calculation
        """
        print('tags')
        print(tags)
        print()
        #get bands from these files
        gw_bands = self.get_bands(tags=tags,path_kpoints=path_kpoints,type_calc=('gw',))[1]

        #plot the bands
        return gw_bands.plot(show=False)

    def plot_bse(self,tags,cols=(2,),ax=None,png_file=False, **kwargs):
        """
        Use this function to plot the absorption spectrum calculated using the BSE
        cols: a list of indexes to select which columns from the file to plot

        Example:
            a.plot_bse('eps_q1',cols=(2,))

            Will plot only files with 'eps_q1' in their filename (absorption spectra)
            Will plot the second column (absorption spectra)


            a.plot_bse(('eps_q1,'FFTGvecs'),cols=(2,))
            Will plot only files with 'eps_q1' in their filename (absorption spectra)
            and for FFTGvecs variable

        !!!!!
        !!!!! Problems here. I have removed there reference to data ["data"]

        """
        import matplotlib.pyplot as plt
        if ax is None:
            standalone = True
            ax = plt.gca()
        else:
            standalone = False
        plot = False

        colors = self.get_colors(tags)

        n=0
        x_key = kwargs.pop('x_key','E/ev[1]')
        y_key = kwargs.pop('y_key','EPS-Im[2]')

        for k in sorted(self.jsonfiles.keys()):
            for filename in list(self.jsonfiles[k]["files"].keys()):
                if all(i in filename for i in tags):
                    # I prefer to work directly with the dictionary...
                    #data = np.array( self.jsonfiles[k]["files"][filename] )
                    data = self.jsonfiles[k]["files"][filename]
                    #select the color to plot with
                    color = colors[n]
                    n+=1

                    for col in cols:
                        #x = data[:,0]
                        #y = data[:,col-1]
                        x = data[x_key]
                        y = data[y_key]
                        label = filename.split('/')[-1]+" col=%d"%col
                        # Should we clean the label?
                        label=label.replace('.eps_q1_haydock_bse','')
                        label=label.replace('o-','')
                        ax.plot(x,y,label=label,color=color)
                        plot = True
        if plot:
            ax.set_ylabel(r'Im$[\\chi(\omega)]$')
            ax.set_xlabel(r'$\omega$ (eV)')

            ax.legend(frameon=False,loc=1)
            if standalone: plt.show()
        if png_file:
            plt.savefig('%s.png' % label[:8])

        return ax

    def plot_spectral_function(self,tags):
        """
        Plot the spectral function
        """
        if isinstance(tags,str):
            tags = (tags,)

        ax = plt.axes([0.1, 0.1, .7, .7])
        for json_filename in sorted(self.jsonfiles.keys()):
          for output_filename in self.jsonfiles[json_filename]["data"]:
            if all(i in output_filename for i in tags):
               data = np.array( self.jsonfiles[json_filename]["data"][output_filename] )
               plt.title('Spectral function as a function of Temperature')
               plt.plot(data[:,0],data[:,2],'-',label=output_filename)
        box = ax.get_position()
        ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])

        ax.legend(loc='center left', bbox_to_anchor=(1, 0.5), prop={'size':8})
        plt.show()

    def print_inputfiles(self):
        """
        Print all the inputfiles from all the json files
        """
        #iterate over the json files
        for k in list(self.jsonfiles.keys()):
            print("jsonfile: ", k)

            #get the jsonfile
            jsonfile = self.jsonfiles[k]

            for inputfile,content in list(jsonfile['inputfile'].items()):
                print("inputfile:", inputfile)
                y = YamboIn(filename=None)
                y.arguments = content["arguments"]
                y.variables = content["variables"]
                print(y+'\n')

    def plot_gw_all_kpoints_convergence(self,tag=None):
        '''
        Function to plot the GW-QPs energies of all k-points
        for a given tag
        Please test
        '''
        import matplotlib.pyplot as plt
        # 1. Find the json files with the tag given: FFTGvecs, BndsRnXp, etc.

        tag_list = []

        for word in self.jsonfiles.keys():
            if tag in word:
               tag_list.append(word.replace('.json','') )

        ntags = len(tag_list)

        cmap = plt.get_cmap('rainbow')
        colors=[cmap(i) for i in np.linspace(0,1,ntags)]

        bands_tag = []

        # 2. Get the bands of all keys
        for it in range(ntags):
            bands_tag.append(self.get_bands(tags=tag_list[it],type_calc=('gw',))[1])
            
        # 3. Get the bands of all keys
        fig = plt.figure()
        ax = fig.add_subplot(1,1,1)
        for it in range(ntags):
            bands_tag[it].plot_ax(ax,color_bands=colors[it],c_label=tag_list[it],legend=True)

        plt.show()

    def __str__(self):
        lines = []; app = lines.append
        app(marquee(self.__class__.__name__))
        for json_file in self.jsonfiles:
            app("%s"%json_file)
            for f in self.jsonfiles[json_file]['files']:
                app("\t%s"%f)
        return "\n".join(lines)



import numpy as np
from yambopy.dbs.excitondb import YamboExcitonDB
from yambopy.dbs import excitondb
from yambopy.dbs.latticedb import YamboLatticeDB
from yambopy.dbs.wfdb import YamboWFDB
import os

iqpt = 1 # qpt index of exciton
calc_path = '.'
BSE_dir = 'GW_BSE'

# load lattice db
lattice = YamboLatticeDB.from_db_file(os.path.join(calc_path, 'SAVE','ns.db1'))

# load exciton db
# note in case too many excitons, load only first few with `neigs' flag
# DO NOT forget to include all degenerate states when giving neigs flag !
#
filename = 'ndb.BS_diago_Q%d' % (iqpt)
excdb = YamboExcitonDB.from_db_file(lattice, filename=filename,
                                    folder=os.path.join(calc_path, BSE_dir),
                                    neigs = -1)

#Load the wavefunction database
wfdb = YamboWFDB(path='.', latdb=lattice,
                      bands_range=[np.min(excdb.table[:, 1]) - 1,
                      np.max(excdb.table[:, 2])])

## plot the exciton wavefunction with hole fixed at [0,0,0]
# in a [1,1,1] supercell with 80 Ry wf cutoff. (give -1 to use full cutoff)
# I want to set the degeneracy threshold to 0.01 eV
# For example I want to plot the 3rd exciton, so iexe = 2 (python indexing )
#
excdb.real_wf_to_cube(iexe=2, wfdb=wfdb, fixed_position=[0.0      ,  0.0 ,       0.0],
                    supercell=[1,1,1], degen_tol=0.01, wfcCutoffRy=-1, fix_particle='h')
# fixed_position is in reduced units
# in case, you want to plot hole density by fixing electron, set fix_particle = 'e'

## .cube will be dumped and use vesta to visualize it !




"""
Tutorial for YamboElectronPhononDB and LetzElphElectronPhononDB.

Electron-phonon matrix element reading and plotting

EDIT the path below to point to the yambo SAVE folder.
"""
save_path='.'
from yambopy import YamboLatticeDB,YamboElectronPhononDB,LetzElphElectronPhononDB
from yambopy.units import ha2ev,bohr2ang
from yambopy.plot.plotting import shifted_grids_2D,BZ_hexagon
import numpy as np
import matplotlib.pyplot as plt

def get_plottable(elph,b0,b1,units='eV'):
    """
    Prepare the quantity to be plotted, which is:

    :: sqrt(  1/4 sum_{n1,n2} |g_{n1n2}^M_def (Q_def,k)|^2 )

    - M_def (mode) and Q_def (Q) already selected in input elph
    - sum factor is 1/4 because we assume only 2 consecutive bands, either b0=0,b1=1 or b0=2,b1=3
    - expression is converted from Rydberg to meV
    """
    if units=='meV':   dim = 1.
    elif units=='eV':  dim = 1000.
    elif units=='Ha':  dim = ha2ev*1000.
    elif units=='Ry':  dim = ha2ev/2.*1000.
    else: raise ValueError("elph units not recognized")

    elph2plot = np.sum( elph[:,b0:b1+1,b0:b1+1], axis=(1,2) )
    return np.sqrt(elph2plot/4.)*dim

def plot_2D_elph(pts,data,rlat=None,plt_cbar=False,**kwargs):
    """
    This is a custom version of the `plot_elph` function extracted from
    YamboElectronPhononDB. You can make your own plotting functions.
    """
    fig, ax = plt.subplots(1,1)

    # Draw 2D BZ hexagonal borders
    ax.add_patch(BZ_hexagon(rlat))

    if plt_cbar:
        if 'cmap' in kwargs.keys(): color_map = plt.get_cmap(kwargs['cmap'])
        else:                       color_map = plt.get_cmap('viridis')

    # Reproduce plot also in adjacent BZs
    BZs = shifted_grids_2D(pts,rlat)
    for pts_s in BZs: plot=ax.scatter(pts_s[:,0],pts_s[:,1],c=data,**kwargs)

    if plt_cbar: cbar = fig.colorbar(plot)

    plt.gca().set_aspect('equal')

    lim = np.linalg.norm(rlat[1])
    lim = 0.85*np.linalg.norm(rlat[0])
    ax.set_xticks([])
    ax.set_yticks([])
    ticks = [-0.15,-0.1,-0.05,0,0.05,0.1,0.15]
    ticklabels = [ "{:.1f}".format(tick*2.*np.pi/bohr2ang) for tick in ticks ]
    ax.set_xticks(ticks)
    ax.set_yticks(ticks)
    ax.set_xticklabels(ticklabels)
    ax.set_yticklabels(ticklabels)
    ax.set_xlim(-lim,lim)
    ax.set_ylim(-lim,lim)
    ax.set_xlabel(r'$q_x$ ($\mathring{A}^{-1}$)',fontsize=12)
    ax.set_ylabel(r'$q_y$ ($\mathring{A}^{-1}$)',fontsize=12)
    #cbar.ax.set_ylabel(cbarlabel,fontsize=12)

if __name__ == "__main__":
    """
    Main part of the script
    """
    # Which database to read
    elphdb        = 'lelphc'
    elphdb        = 'yambo'
    
    # In which momentum space to plot
    kspace_Plot    = True # k-space at fixed q 
    qspace_Plot    = True # q-space at fixed k

    # MoS2 details
    # * Mode A_1^\prime --> Index 7 (penultimate mode)
    # * v bands --> Index 0 [24], 1 [25]
    # * c bands --> Index 2 [26], 3 [27]

    # Customly chosen matrix elements
    i_v, i_vp = [0,1] #We choose the two valence bands (spin-orbit split at K/K')
    i_nu = 7          # A1^\prime phonon mode 
    iK_qspace = 34    # indx of K-pt in 12x12x1 BZ for q-space (phonon momenta)
    iK_kspace = 34    # indx of K-pt in 12x12x1 BZ for k-space (electron momenta)

    # Create "lattice" object from ns.db1 db
    ylat = YamboLatticeDB.from_db_file(filename=save_path+'/SAVE/ns.db1')
    # Restrict plot two q=0 plane (no effect in true 2D)
    indx2D=(ylat.car_kpoints[:,2]==0.).nonzero()[0]
    # In yambo, k and q grids coincide
    ptgrid = ylat.car_kpoints[indx2D]

    # Create "elphon" object by reading:
    # (i) either the ndb.elph_gkkp* databases inside the yambo SAVE
    # (ii) or the ndb.elph database from lelphc
    if elphdb=='lelphc':
        
        elph = LetzElphElectronPhononDB(save_path+'/ndb.elph')
        print(elph)

        # Print info on how to use this class
        print(elph.__doc__)

        el_ph_mat_squared = elph.gkkp_sq[:,:,:,0,:,:]
        units = 'Ry'

    elif elphdb=='yambo':
        
        elph = YamboElectronPhononDB(ylat,folder_gkkp=save_path+'/SAVE',save=save_path+'/SAVE')
        print(elph)

        # Print info on how to use this class
        print(elph.__doc__)
    
        el_ph_mat_squared = elph.gkkp_sq
        units= 'Ha'
    
    else:
        print(f"elphdb string must be 'yambo' or 'lelphc', you have {elphdb}")
        exit()

    # We plot the average of \sqrt{|g_{vv'm}(k,q)|^2} over v and v', output in meV    
    # We select a specific Q-point to plot |g(K)| in kspace
    g_of_k = get_plottable( el_ph_mat_squared[iK_qspace,:,i_nu,:,:],i_v,i_vp,units=units )

    # We select a specific K-point to plot |g(Q)| in qspace
    g_of_q = get_plottable( el_ph_mat_squared[:,iK_kspace,i_nu,:,:],i_v,i_vp,units=units )
   
    # Plots are customisable as needed using matplotlib
    if kspace_Plot:

        plot_2D_elph(ptgrid,g_of_k,rlat=ylat.rlat,plt_cbar=True,\
                     marker='H',s=700,cmap='magma')
        plt.title(r'$|g(k)|$ (meV)',fontsize=8)
        plt.savefig('g_kspace.png',dpi=200)
        plt.show()

    if qspace_Plot:

        plot_2D_elph(ptgrid,g_of_q,rlat=ylat.rlat,plt_cbar=True,\
                     marker='H',s=700,cmap='magma')
        plt.title(r'$|g(q)|$ (meV)',fontsize=8)
        plt.savefig('g_qspace.png',dpi=200)
        plt.show()




# Author: 
# Tests for the yambopy library
# 
#
from __future__ import print_function
import unittest
import sys
import os
import argparse
import subprocess
import filecmp
from yamboparser import YamboFile, YamboFolder

folder = os.path.join(os.path.dirname(__file__),'..','..','yambopy','data','refs','parser')

class TestFolder(unittest.TestCase):
     
    def test_folder_list(self):
        fold = YamboFolder(os.path.join(folder,'t2_parse_qps/'))
        assert len (fold.yambofiles)==5


class TestFileT1(unittest.TestCase):

    def test_qp_parsing(self):
        fl = YamboFile('o-GW_run.10.720.qp',os.path.join(folder,'t1_errors_warnings'))
        assert list(fl.data.keys()) == ['1','7','13','40']  # K points are the keys in default mode
        assert  fl.type == 'output_gw'

    def test_zip_tags_parsing(self):
        parse_kwargs = {'zip_tags':True,'dummy_kwarg':False} #dummy kwarg to check that irrelevant kwargs don't break things
        fl = YamboFile('o-GW_run.10.720.qp',os.path.join(folder,'t1_errors_warnings'),**parse_kwargs)
        assert list(fl.data.keys()) == ['K-point','Band','Eo','E-Eo','Sc|Eo'] # columns of file are keys in zip mode

    def test_l_parsing(self):
        fl = YamboFile('l-GW_run.8.480_em1d_ppa_HF_and_locXC_gw0_rim_cut_CPU_1',os.path.join(folder,'t1_errors_warnings'))
        assert  not fl.data 
        assert len(fl.warnings) ==1
        assert len(fl.errors) == 1
        assert  fl.type == 'log'

    def test_r_parsing(self):
        fl = YamboFile('r-GW_run.8.480_em1d_ppa_HF_and_locXC_gw0_rim_cut',os.path.join(folder,'t1_errors_warnings'))
        assert fl.type=='report'
        assert fl.kpoints
        assert not fl.data

class TestFileT2(unittest.TestCase):

    def test_qp_parsing(self):
        fl = YamboFile('o-yambo.qp',os.path.join(folder,'t2_parse_qps'))
        assert  fl.type == 'output_gw'

    def test_l_parsing(self):
        fl = YamboFile('l-yambo_em1d_HF_and_locXC_gw0',os.path.join(folder,'t2_parse_qps'))
        assert  fl.type == 'log'

    def test_r_parsing(self):
        fl = YamboFile('r-yambo_em1d_life',os.path.join(folder,'t2_parse_qps'))
        assert fl.type=='report'

        fl = YamboFile('r-yambo_em1d_HF_and_locXC_gw0',os.path.join(folder,'t2_parse_qps'))
        assert fl.type=='report'

    def test_ndb_qp_parsing(self):
        fl = YamboFile('ndb.QP',os.path.join(folder,'t3_parse_netcdf'))
        print("fl type", fl.type)
        assert fl.type=='netcdf_gw'

    def test_ndb_hf_parsing(self):
        fl = YamboFile('ndb.HF_and_locXC',os.path.join(folder,'t3_parse_netcdf'))
        print("fl type", fl.type)
        assert fl.type=='netcdf_hf'

if __name__ == "__main__":

    #t1_errors_warnings
    suite = unittest.TestLoader().loadTestsFromTestCase(TestFileT1)
    unittest.TextTestRunner(verbosity=2).run(suite)

    #t2_parse_qps
    suite = unittest.TestLoader().loadTestsFromTestCase(TestFileT2)
    unittest.TextTestRunner(verbosity=2).run(suite)




# 
# License-Identifier: GPL
#
# Copyright (C) 2024 The Yambo Team
#
# Authors: HM, AM-S, JC-V, FP
# First verion by HM and AM-S (2017), revised and expanded by JC-V (2025)
#
# This file is part of the yambopy project
#
#
from qepy import *
import numpy as np
from sys import stdout
import h5py

class Unfolding():

    def __init__(self,prefix_pc,prefix_sc,path_pc='.',path_sc='.',spin="none",band_min=0,sc_rotated=False,compute_projections=True):
        """ 
        Initialize the structure with data from the primitive cell and the supercell and then compute the projection of the supercell band structure
        taking the primitive cell band structure as a reference

        === Usage and variables ===

        >> Unfold = Unfolding(prefix_pc=db_prefix_pc,path_pc=db_path_pc,prefix_sc=db_prefix_sc,path_sc=db_path_sc,spin="noncol",band_min = 0,sc_rotated=False,compute_projections=True)
        >> Unfold.plot_eigen_ax(ax,path=db_path_kpoints_sc,ylim = (ylim_min,ylim_max))
 
        Input:
        :: prefix_pc(sc) is the prefix of the primitive cell (supercell) QE save
        :: path_pc(sc) is the path of the primitive cell (supercell) along the Brillouin zone
        :: spin is equal to "none" or "noncol" for nspin = 1 (noSOC) and nspin 4 (SOC) calculations according to QE format
        :: band min determines the number of computed bands. If band_min = 0, it computes the projection of all the bands. If band_min != 0, it computes the projection of the nbands-band_min bands
        :: sc_rotated determines the rotation matrix if necessary. If sc_rotated = False, it takes the matrix identity. If sc_rotated = (3x3) matrix, it takes the rotation matrix defined by the user
        :: compute_projections determines whether to calculate projections or not. If compute_projections = True, it calculates the projections and save them in a .npy file.
           If compute_projections = False, it loads the previous .npy file.
        
        Output:
        :: projections.npy file to plot the EBS with an external script
        :: Plot with the EBS of the supercell taking as a reference the defined primitive cell
        """

        print("=== Initializing the data ===")

        # Prefix and path of primitive cell and supercell
        self.prefix_pc = prefix_pc  
        self.prefix_sc = prefix_sc
        self.path_pc   = path_pc 
        self.path_sc   = path_sc
       
        # Reading primitive cell and supercell database from QE
        pc_xml = PwXML(prefix=self.prefix_pc,path=self.path_pc) 
        sc_xml = PwXML(prefix=self.prefix_sc,path=self.path_sc)

        # Number of kpoints of the primitive cell and supercell
        self.nkpoints_pc = pc_xml.nkpoints 
        self.nkpoints_sc = sc_xml.nkpoints 

        # List of kpoints of the supercell 
        self.kpoints = sc_xml.kpoints  

        # Number of bands of the primitive cell and supercell
        self.nbands_pc = pc_xml.nbands  
        self.nbands_sc = sc_xml.nbands 

        # This indicates the code to compute the projections of nbands - band_min bands
        self.band_min  = band_min 

        # Condition to control the number of computed bands  
        if self.band_min > self.nbands_sc:
           raise Exception("Minimum of bands larger than total number of bands")

        # Reciprocal lattice of the primitive cell and supercell in cartesian coordiantes
        self.rcell_pc = array(pc_xml.rcell)/pc_xml.cell[0][0] 
        self.rcell_sc = array(sc_xml.rcell)/sc_xml.cell[0][0]
   
        # Eigenvalues of the primitive cell and supercell
        self.eigen_pc = array(pc_xml.eigen1) 
        self.eigen_sc = array(sc_xml.eigen1) 

        # Format to save Miller indices
        format_string = "%12.4lf %12.4lf %12.4lf" 
        n_decs = 8 

        # Distance of the k-path
        kpoints_dists = calculate_distances(self.kpoints) 
       
        # Condition to use a rotation matrix if needed 
        if sc_rotated is False:
            self.rot = np.array([ [1.0,0.0,0.0], [0.0,1.0,0.0], [0.0,0.0,1.0] ])

        elif isinstance(sc_rotated, np.ndarray) and sc_rotated.shape == (3,3):
            self.rot = sc_rotated

        else:
            raise Exception("Input for sc_rotated must be False or a 3x3 matrix")

        # Array to save the projections
        self.projection = zeros([self.nkpoints_sc,self.nbands_sc-self.band_min]) 

        print("=== Data initialized successfully ===")

        print("=== Calculation of the projections  ===")

        # Condition to compute the projections or to use the saved ones  
        if compute_projections == True:

           # Loop along the kpoints of the supercell  
           for ik in range(self.nkpoints_sc):
               load(ik,self.nkpoints_sc)

               # Loading the data of each kpoint of the primitive cell and supercell 
               f_pc = h5py.File('%s/%s.save/wfc%01d.hdf5' % (self.path_pc,self.prefix_pc,(ik + 1)), 'r') 
               f_sc = h5py.File('%s/%s.save/wfc%01d.hdf5' % (self.path_sc,self.prefix_sc,(ik + 1)), 'r') 

               # Dimension of the Miller indices
               self.ng_pc = int(f_pc.attrs['igwx'])
               self.ng_sc = int(f_sc.attrs['igwx']) 

               # Creating a dictionary to collect key-value pairs
               g_sc = dict()
               g_sc_int = dict()
           
               # Loading Miller indices
               mill_sc = f_sc['MillerIndices']
               mill_pc = f_pc['MillerIndices']
        
               # Loop along the number of Miller indices of the super cell
               for ig in arange(self.ng_sc):
              
                   # Definition of Miller indices in order to reconstruct the k-points (to rotate it if neccessary) 
                   # and to save them in the dictionary associated to an integer
                   h,k,l = mill_sc[ig,0], mill_sc[ig,1], mill_sc[ig,2]
                   g_sc_int[(int(h),int(k),int(l))] = ig
                   w = h*self.rcell_sc[:][0] + k*self.rcell_sc[:][1] + l*self.rcell_sc[:][2]
                   w = dot(self.rot,w)
                   w = np.around(w, decimals = n_decs) + array([0,0,0])
                   w = format_string % (abs(w[0]), abs(w[1]), abs(w[2]))
                   g_sc[w] = ig
                   
               # To check in the following loop if the k-point of the super cell is found in the primitive cell    
               g_contain = [0]*self.ng_pc

               # Loop along the number of Miller indices of the primitive cell
               for ig in arange(self.ng_pc):

                   # Definition of Miller indices in order to reconstruct the k-points
                   h,k,l = mill_pc[ig,0], mill_pc[ig,1], mill_pc[ig,2]
                   w = h*self.rcell_pc[:][0] + k*self.rcell_pc[:][1] + l*self.rcell_pc[:][2]
                   w = np.around(w, decimals = n_decs) + array([0,0,0])
                   w = format_string % (abs(w[0]), abs(w[1]), abs(w[2]))

                   # Checking if the k-point in the supercell is found in the primitive cell,
                   # if missing, the projection will be wrong.
                   try:
                       g_contain[ig] = g_sc[w]
                   except KeyError:
                       print("Missing k-point %d" % ig)
           

               # Condition to read the eigenvectors for nspin = 1 or nspin = 4 in QE (To implement spin == col, i.e., nspin = 2 in QE)
               if spin == "none" or spin == "noncol":
              
                  # Loading eigenvectors of the super cell
                  evc_sc = f_sc['evc']
                  eivecs = []

                  # Loop along the number of bands indicated by the user to be computed
                  for ib in range(self.band_min,self.nbands_sc):
                      eivec = evc_sc[ib, :]
                      
                      # Rewriting the eigenvectors to manage them properly
                      eivec_complex = [complex(eivec[i], eivec[i+1]) for i in range(0, len(eivec), 2)]

                      eivecs.append(eivec_complex)
                        
                      # Defining specifically ib == 0 since it presents one component instead of two as the rest of values
                      if ib==0:
                         x = 0.0
                         for ig in range(self.ng_sc):
                             x += eivecs[-1][ig]*eivecs[-1][ig].conjugate()
 
               # Condition to compute the projections for nspin = 1 or spin = 4 in QE (To implement spin == col, i.e., nspin = 2 in QE)
               if spin == "none" or spin == "noncol":

                  # Loop along the number of bands indicated by the user to be computed
                  for ib in range(self.nbands_sc-self.band_min): 
                      x = 0.0
                      for ig in range(self.ng_pc): 

                          # Computing the projection between the primitive cell and the super cell
                          x += eivecs[ib][g_contain[ig]]*(eivecs[ib][g_contain[ig]].conjugate())

                      # If the value is less than a threshold, the projection is set to zero (to avoid ficticious points when plotting)
                      if abs(x) < 1e-4:

                         self.projection[ik,ib] = 0.0
                     
                      else:

                         self.projection[ik,ib] = abs(x)
               
               # Saving the data to avoid recomputing the projections twice
               np.save('projections',self.projection)


        # Loading the projections when already computed
        if compute_projections == False:

           print(" * Loading projections ...  ")

           try:

              self.projection = np.load('projections.npy')

           except FileNotFoundError:

              print(f"Error: the file projections.npy does not exits")
              raise

        return print("=== Projections calculated successfully ===")

    def plot_eigen_ax(self,ax,path=[],xlim=(),ylim=()):
        """
        Provisional plot function for quick visualization of the data. Useful for small calculations. 
        For large calculations is more useful loading the .npy and plot it with another script.
        """

        print(" * Plotting EBS ...  ")

        # Getting the data of the defined path
        if path:
            if isinstance(path,Path):
                path = path.get_indexes()
            ax.set_xticks( *list(zip(*path)) )
        ax.set_ylabel('E (eV)')

        # Computing the distance of the path 
        kpoints_dists = calculate_distances(self.kpoints)

        # Defining labels
        ticks, labels = list(zip(*path))
        ax.set_xticks([kpoints_dists[t] for t in ticks])
        ax.set_xticklabels(labels)
        ax.set_ylabel('E (eV)')

        # Plotting high-symmetry vertical lines
        for t in ticks:
            ax.axvline(kpoints_dists[t],c='k',lw=2)
        ax.axhline(0,c='k',lw=1)

        # Plotting the band for the primitive cell
        for ib in range(self.nbands_pc):
           ax.plot(kpoints_dists,self.eigen_pc[:,ib],'k--',lw=0.5)

        # Plotting the projection of the super cell 
        for ib in range(self.nbands_sc-self.band_min):
           #ax.plot(kpoints_dists,self.eigen_sc[:,ib],'darkgreen',lw=0.2)
           ax.scatter(kpoints_dists,self.eigen_sc[:,ib+self.band_min],s=self.projection[:,ib]*5,color='navy',edgecolor = None)

        # Establishing x and y axis limits 
        if xlim: ax.set_xlim(xlim)
        if ylim: ax.set_ylim(ylim)

        return print("=== EBS calculated successfully  ===")

# Load bar
def load(x,n):
    bar_length = 100
    x+=1
    ratio = x/float(n)
    c = int(ratio * bar_length)
    stdout.write(" * Computing projections of the supercell onto the primitive cell ["+"="*c+" "*(bar_length-c)+"] %03.3f%%" % (ratio*100))
    if (x==n): stdout.write("\n")
    stdout.flush()
    stdout.write("\r")




# Copyright (C) 2018 Henrique Pereira Coutada Miranda, Alejandro Molina-Sanchez
# All rights reserved.
#
# This file is part of yambopy
#
#
from __future__ import print_function
import subprocess
from .scheduler import Scheduler

class Oar(Scheduler):
    """
    Class to submit jobs through the OAR scheduler

    ``_vardict`` states the default assignement of the nodes and cores variables
    from the schduler class to the variables needed in this class
    """
    _vardict = {"cores":"core",
                "nodes":"nodes"}
                
    def initialize(self):
        self.get_vardict()
        args = self.arguments
        if self.get_arg("besteffort"): args.append("-t besteffort")
        if self.get_arg("idempotent"): args.append("-t idempotent")
        if self.get_arg("bigmem"): args.append("-t bigmem")
        
        if self.name:  args.append("-n \"%s\""%self.name)
        dependent = self.get_arg("dependent")
        if dependent: args.append("-a %d"%dependent)
        
        resources_line = self.get_resources_line()
        if resources_line:
            args.append(resources_line)
        
    def get_resources_line(self):
        """
        get the the line with the resources
        """
        s = "-l "
        if self.nodes: s += "%s=%d"%(self.vardict['nodes'],self.nodes)
        if self.nodes and self.cores: s += "/"
        if self.cores: s += "%s=%d"%(self.vardict['cores'],self.cores)
        if self.nodes or self.cores: s+= ","
        s += "walltime=%s"%self.walltime
        return s
    
    def get_script(self):
        """
        get a .pbs file to be submitted using qsub
        qsub <filename>.pbs
        """
        s = '#!/bin/bash\n'
        s += "\n".join(["#OAR %s"%s for s in self.arguments])+'\n'
        s += self.get_commands()
        return s
        
    def get_bash(self):
        """
        get a bash command to submit the job
        """
        command  = "oarsub "
        command += " \\\n".join(self.arguments)+" "
        command += "\"%s\""%self.get_commands().replace("'","\"").replace("\"","\\\"")
        return command
        
    def __str__(self):
        """
        create the string for this job
        """
        return self.get_script()

    def run(self,silent=True,dry=False):
        """
        run the command
        arguments:
        dry - only print the commands to be run on the screen
        """
        command = self.get_bash()
        
        if dry:
            print(command)
        else:
            p = subprocess.Popen(command,stdout=subprocess.PIPE,stderr=subprocess.PIPE,shell=True)
            self.stdout,self.stderr = p.communicate()
            
            #check if there is stderr
            if self.stderr: raise Exception(self.stderr)
            
            #check if there is stdout
            if not silent: print(self.stdout)
                
            #get jobid
            for line in self.stdout.split('\n'):
                if 'OAR_JOB_ID' in line:
                    self.jobid = int(line.strip().split('=')[1])
            print("jobid:",self.jobid)



import os
import re
from subprocess import Popen, PIPE
from yambopy.env import yambopyenv
from yambopy.tools.duck import isstring

class YamboIO(object):
    """
    This class provides input/output capabilities to yambopy.

    Currently: called by other classes via the msg function, 
               it opens/closes a LOG file and fills it with the messages 

    Example of use:

        .. code-block:: python

            yf = YamboIO(print_to_shell=False)
            yf.IO_start()
            yf.msg('message to print')
            yf.IO_close()
    """
    def __init__(self,out_name='yambopy.log',out_path='.',print_to_shell=True):
        self.fpath = '%s/%s'%(out_path,out_name)
        self.print_to_shell = print_to_shell
        self.s = ""

    def IO_start(self):
        self.f = open(self.fpath,'w')

    def IO_close(self):
        self.f.close()

    def msg(self,msg_string):
        self.f.write('%s\n'%msg_string)
        self.s += '%s\n'%msg_string
        if self.print_to_shell: print(msg_string)   

    def __str__(self):
        s = self.s
        return s
